"""
Parse Cannalysis CoAs
Copyright (c) 2022 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 8/2/2022
Updated: 9/22/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse Cannalysis CoA PDFs.

Data Points:

    - analyses
    ✓ {analysis}_method
    ✓ {analysis}_status
    ✓ date_collected
    ✓ date_received
    ✓ date_tested
    ✓ date_produced
    ✓ batch_size
    ✓ lab_id
    - lab_results_url
    ✓ metrc_lab_id
    ✓ metrc_source_id
    ✓ product_name
    ✓ product_type
    ✓ results
    ✓ results_hash
    ✓ sample_hash
    ✓ sample_size
    ✓ total_cannabinoids
    ✓ total_cbd
    ✓ total_thc
    ✓ total_terpenes

FIXME:

    - [ ] Occasional unknown error: `list index out of range`

"""
# Standard imports.
from datetime import datetime
import json
import os
import re
from typing import Any, Optional
import warnings

# External imports.
import pandas as pd
import pdfplumber
from PIL import Image

# Internal imports.
from cannlytics import __version__
from cannlytics.data.data import (
    create_hash,
    create_sample_id,
    find_first_value,
)
from cannlytics.firebase import (
    create_id,
    create_short_url,
    get_file_url,
    initialize_firebase,
    upload_file,
)
from cannlytics.utils.constants import (
    ANALYSES,
    ANALYTES,
    STANDARD_FIELDS,
)
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
)

# Convert `DecompressionBombWarning` into an error.
warnings.simplefilter('error', Image.DecompressionBombWarning)

# It is assumed that the lab has the following details.
CANNALYSIS =  {
    'coa_algorithm': 'cannalysis.py',
    'coa_algorithm_entry_point': 'parse_cannalysis_coa',
    'url': 'www.cannalysis.com',
    'lims': 'Cannalysis',
    'lab': 'Cannalysis',
    'lab_license_number': 'C8-0000012-LIC',
    'lab_image_url': 'https://www.cannalysis.com/img/img.c5effdd3.png',
    'lab_address': '1801 Carnegie Ave, Santa Ana CA 92705',
    'lab_street': '1801 Carnegie Ave',
    'lab_city': 'Santa Ana',
    'lab_county': 'Orange',
    'lab_state': 'CA',
    'lab_zipcode': '92705',
    'lab_latitude': 33.712190,
    'lab_longitude': -117.844650,
    'lab_phone': '949-329-8378',
    'lab_email': 'support@cannalysislabs.com',
    'lab_website': 'www.cannalysis.com',
}

# It is assumed that the CoA has the following parameters.
# Dimensions are percentages (x0, y0) to (x1, y1).
CANNALYSIS_COA = {
    'coa_distributor_area': [0, 0.66, 0.35, 0.725],
    'coa_producer_area': [0, 0.725, 0.35, 0.8],
    'coa_page_area': [
        [0, 0.1, 0.5, 0.9],
        [0.5, 0.1, 1.0, 0.9],
    ],
    'coa_sample_details_area': [
        [0, 0.15, 0.35, 0.9],
        [0.35, 0.15, 1.0, 0.9],
    ],
    'coa_image_area': [0.049, 0.179, 0.32, 0.3876],
    'coa_result_fields': [
        'name',
        'value',
        'lod',
        'loq', 
        'limit',
        'status',
    ],
    'coa_skip_fields': [
        'ADDITIONAL',
        'Total',
        'undergone a calculation',
        'Individual Analyte',
        '   ',
        'AS',
        'Batch:',
        'ACCORDANCE',
        'criteria',
        'identicalorsimilarproducts',
        'bstractedinanymanner',
        'Any violation of these conditions',
        'All LQC samples required',
        'acceptance criteria',
    ],
}


def parse_cannalysis_coa(
        parser,
        doc: Any,
        cleanup: Optional[bool] = True,
        resolution: Optional[int] = 300,
        temp_path: Optional[str] = '/tmp',
        coa_pdf: Optional[str] = None,
        **kwargs,
    ) -> Any:
    """Parse a Cannalysis CoA PDF.
    Args:
        parser (CoADoc): A CoADoc parsing client.
        doc (str or PDF): A PDF file path or pdfplumber PDF.
        cleanup (bool): Whether or not to remove the files generated
            during OCR, `True` by default (optional).
        temp_path (str): A temporary directory to use for OCR.
        resolution (int): The resolution of rendered PDF images,
            300 by default (optional).
        coa_pdf (str): A filename to use for the `coa_pdf` field (optional).
    Returns:
        (dict): The sample data.
    """

    # Read the PDF.
    obs = {}
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = coa_pdf or doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = coa_pdf or report.stream.name.replace('\\', '/').split('/')[-1]

    # Get the QR code from the last page.
    # Note: After OCR, QR code decoding raises DecompressionBombWarning.
    try:
        obs['lab_results_url'] = parser.find_pdf_qr_code_url(
            report,
            resolution=resolution,
        )
    except Image.DecompressionBombWarning:
        obs['lab_results_url'] = ''

    # Get the lab specifics.
    coa_parameters = CANNALYSIS_COA
    standard_result_fields = coa_parameters['coa_result_fields']
    skip_fields = coa_parameters['coa_skip_fields']

    # Get the front page and page dimensions.
    front_page = report.pages[0]
    w, h = front_page.width, front_page.height

    # Get the distributor area.
    x0, y0, x1, y1 = tuple(coa_parameters['coa_distributor_area'])
    coa_distributor_area = (x0 * w, y0 * h, x1 * w, y1 * h)

    # Get the producer area.
    x0, y0, x1, y1 = tuple(coa_parameters['coa_producer_area'])
    coa_producer_area = (x0 * w, y0 * h, x1 * w, y1 * h)

    # Get the page areas.
    coa_page_area = []
    for dimensions in coa_parameters['coa_page_area']:
        x0, y0, x1, y1 = tuple(dimensions)
        area = (x0 * w, y0 * h, x1 * w, y1 * h)
        coa_page_area.append(area)    

    # Get the sample details areas.
    coa_sample_details_area = []
    for dimensions in coa_parameters['coa_sample_details_area']:
        x0, y0, x1, y1 = tuple(dimensions)
        area = (x0 * w, y0 * h, x1 * w, y1 * h)
        coa_sample_details_area.append(area)

    # Get the distributor data based on page area.
    crop = front_page.within_bbox(coa_distributor_area)
    lines = crop.extract_text().split('\n')[1:]
    lines = [x.replace('  ', ' ').strip() for x in lines]
    obs['distributor_address'] = ' '.join([lines[0], lines[1]])
    obs['distributor_license_number'] = lines[-1].replace('License:', '').strip()

    # Get the producer data based on page area.
    crop = front_page.within_bbox(coa_producer_area)
    lines = crop.extract_text().split('\n')[1:]
    lines = [x.replace('  ', ' ').strip() for x in lines]
    obs['producer_address'] = ', '.join([lines[0], lines[1]])
    obs['producer_license_number'] = lines[-1].replace('License:', '').strip()

    # Optional: Is there any way to identify the `producer` and
    # `distributor` with their license numbers? Query Cannlytics API?

    # Get the image data, if the PDF has gone through OCR, then cut out
    # the image first. Then save the image to the cloud, get the file
    # URL, create a short URL to use as the `image_url`.
    # TODO: Resize the image if it is too large?
    # TODO: Turn this into a function?
    # try:
    #     image_data = parser.get_pdf_image_data(
    #         front_page,
    #         image_index=1,
    #         resolution=resolution,
    #     )
    # except IndexError:
    #     area = tuple(coa_parameters['coa_image_area'])
    #     image_data = parser.get_pdf_image_data(
    #         front_page,
    #         bbox=area,
    #         resolution=resolution,
    #     )
    # image_id = obs.get('metrc_lab_id', create_id())
    # image_ref = f'public/data/lab_results/{image_id}/image.jpg'
    # api_key = os.environ['FIREBASE_API_KEY']
    # bucket = os.environ['FIREBASE_STORAGE_BUCKET']
    # project_name = os.environ['FIREBASE_PROJECT_ID']
    # # FIXME:
    # initialize_firebase()
    # upload_file(image_ref, data_url=image_data, bucket_name=bucket)
    # file_url = get_file_url(image_ref, bucket_name=bucket)
    # # short_url = create_short_url(api_key, file_url, project_name)
    # short_url = file_url
    # obs['image_url'] = short_url

    # Check if OCR is required then re-run the routine.
    line = lines[0]
    if re.match(r'^\(cid:\d+\)', line):
        temp_file = f'{temp_path}/ocr-coa.pdf'
        parser.pdf_ocr(
            report.stream.name,
            temp_file,
            temp_path=temp_path,
            resolution=resolution,
            cleanup=cleanup,
        )
        return parse_cannalysis_coa(
            parser,
            report,
            cleanup=cleanup,
            resolution=resolution,
            temp_path=temp_path,
            coa_pdf=report.stream.name,
            **kwargs,
        )

    # Get the sample details.
    analyses = []
    for area in coa_sample_details_area:
        crop = front_page.within_bbox(area)
        lines = crop.extract_text().split('\n')
        analysis = None
        collect = False
        field = None
        for line in lines:

            # Determine any fields.
            line = line.replace('  ', ' ').strip()
            potential_field = STANDARD_FIELDS.get(snake_case(line))
            if potential_field:
                field = potential_field
                collect = True
                continue
            if collect and field:
                if field == 'date_collected':
                    dates = tuple(line.split(', '))
                    obs['date_collected'] = dates[0]
                    obs['date_received'] = dates[1]
                elif field == 'batch_size':
                    dates = tuple(line.split(', '))
                    obs['batch_size'] = dates[0]
                    obs['sample_size'] = dates[1]
                else:
                    obs[field] = line
                collect = False
                field = None
            
            # Collect the data.
            elif collect and analysis:
                obs[f'{analysis}_status'] = line.lower()
                collect = False
                analysis = None

            # Determine analyses.
            # FIXME: Not all analyses (e.g. cannabinoids) are being identified.
            potential_analysis = ANALYSES.get(line)
            if potential_analysis:
                analysis = potential_analysis
                analyses.append(potential_analysis)
                collect = True
                continue

    # Get the product name if it wasn't collected from the details.
    # Hot-fix: OCR misreads "Pass" as "G=>". How does OCR read "Fail"?
    if obs.get('product_name') is None:
        text = front_page.extract_text().split('Testing')[-1].split('\n')
        text = [x.replace('  ', ' ') for x in text if x.strip()]
        obs['product_name'] = text[0].replace('G=>', '').strip()

    # Get all page text, from the 2nd page on.
    results, dates_tested = [], []
    for page_number, page in enumerate(report.pages[1:]):
        for area in coa_page_area:

            # Create cropped area.
            analysis, units = None, None
            crop = page.within_bbox(area)
            lines = crop.extract_text().split('\n')

            # Parse results from each line.
            for line in lines:

                # Strip whitespace for easy line recognition.
                line_text = line.replace(' ', '')

                # Identify the analysis (# FIXME: only works for 1st column).
                # Note: Hot-Fix for foreign matter.
                if 'ANALYSIS' in line or 'FILTH' in line:
                    analysis = line.split(' ANA')[0].title()
                    analysis = ANALYSES.get(analysis)

                # Identify the columns (unnecessary).
                elif 'ANALYTE' in line:
                    # parts = [snake_case(x) for x in line.split(' ')]
                    # columns = [standard_fields.get(x, x) for x in parts]
                    continue

                # Identify the units (# FIXME: only works for 1st column).
                elif 'UNIT' in line:
                    units = line.split(':')[-1].strip()
                    units = line[line.find('(') + 1: line.find(')')]
                    if page_number == 0:
                        units = 'percent' # Convert mg/g to percent.

                # Identify the instrument as the `method`.
                elif 'Instrument:' in line:
                    method = line.split('Instrument: ')[-1].split(' Sample Analyzed:')[0]
                    obs[f'{analysis}_method'] = method
                elif 'Method:' in line:
                    continue

                # Get the totals.
                elif 'TOTAL' in line and '@' not in line:
                    parts = line.split(':')
                    key = snake_case(parts[0].lower())
                    value = parts[1].strip()
                    if '(' in value:
                        value = value[value.find('(') + 1: value.find(')')]
                        value = convert_to_numeric(value, strip=True)
                    obs[key] = value

                # Get the dates tested.
                # FIXME: This is highly suboptimal and causes many errors.
                elif 'SampleApproved' in line_text:
                    date_time = line_text.split('Approved')[-1]
                    date_time = date_time.replace(':', '') \
                        .replace('.', '') \
                        .replace(',', '') \
                        .replace('-', '') \
                        .replace('_', '')
                    date, at = date_time[:10], date_time[10:]
                    try:
                        if len(at) != 4:
                            date_time = pd.to_datetime(date)
                        else:
                            date_time = pd.to_datetime(' '.join([date, at]))
                        dates_tested.append(date_time)
                    except:
                        continue

                # Skip informational rows.
                elif any(s in line or s in line_text for s in skip_fields):
                    continue

                # End at the end of the report.
                elif line_text.startswith('Thisreport'):
                    break

                # Get the results.
                else:

                    # Remove extraneous results.
                    text = line.replace('mg/g', '').replace(' aw ', ' ')
                    text = re.sub('[\(\[].*?[\)\]]|=', '', text)
                    text = text.replace('  ', ' ').strip()

                    # Standardize analytes.
                    first_value = find_first_value(text)
                    name = text[:first_value].replace('\n', ' ').strip()
                    analyte = ANALYTES.get(name, snake_case(name))

                    # Hot-fix to skip non-sensical analytes.
                    if len(analyte) <= 1:
                        continue

                    # Create a result object.
                    result = {
                        'analysis': analysis,
                        'key': analyte,
                        'name': name,
                        'units': units,
                    }

                    # Add keys / values to the result, using standard fields.
                    values = text[first_value:].strip().split(' ')
                    values = [x for x in values if x]
                    try:
                        for i, v in enumerate(values):
                            key = standard_result_fields[i + 1]
                            value = convert_to_numeric(v)
                            if page_number == 0: # Convert mg/g to percent.
                                try:
                                    value = round(0.1 * value, 2)
                                except:
                                    pass
                            result[key] = value
                    except IndexError:
                        continue
                
                    # Hot-fix: Skip (non-)analytes that begin with a digit.
                    # Note: It would be best to improve this logic.
                    if result['key'][0].isdigit():
                        continue

                    # Record the result.
                    results.append(result)

    # Close the report.
    report.close()

    # Get the latest tested at date.
    obs['date_tested'] = max(dates_tested)

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Finish data collection with a freshly minted sample ID.
    # TODO: Fix or make `sample_id` obsolete.
    obs = {**CANNALYSIS, **obs}
    obs['analyses'] = json.dumps(list(set(analyses)))
    obs['results'] = json.dumps(results)
    obs['coa_algorithm_version'] = __version__
    obs['coa_parsed_at'] = datetime.now().isoformat()
    obs['results_hash'] = create_hash(results)
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=obs.get('producer', obs.get('date_tested', 'cannlytics.eth')),
    )
    obs['sample_hash'] = create_hash(obs)
    return obs


# === Test ===
if __name__ == '__main__':

    from cannlytics.data.coas import CoADoc

    # [✓] TEST: Parse a Cannalysis CoA PDF (for a flower product).
    parser = CoADoc()
    doc = '../../../tests/assets/coas/cannalysis/Mothers-Milk-Flower.pdf'
    lab = parser.identify_lims(doc, lims={'Cannalysis': CANNALYSIS})
    assert lab == 'Cannalysis'
    data = parse_cannalysis_coa(parser, doc)
    assert data is not None
    print('Parsed:', doc)

    # [✓] TEST: Parse a Cannalysis CoA PDF (for a concentrate product).
    parser = CoADoc()
    doc = '../../../tests/assets/coas/cannalysis/Citrus-Slurm-Diamonds.pdf'
    lab = parser.identify_lims(doc, lims={'Cannalysis': CANNALYSIS})
    assert lab == 'Cannalysis'
    data = parse_cannalysis_coa(parser, doc)
    assert data is not None
    print('Parsed:', doc)

    # [✓] TEST: Parse a Cannalysis CoA PDF after OCR is applied.
    parser = CoADoc()
    doc = '../../../.datasets/tests/test.pdf'
    lab = parser.identify_lims(doc, lims={'Cannalysis': CANNALYSIS})
    assert lab == 'Cannalysis'
    data = parse_cannalysis_coa(parser, doc)
    assert data is not None
    print('Parsed:', doc)

    # [✓] TEST: Parse a Cannalysis CoA PDF, applying OCR.
    parser = CoADoc()
    doc = '../../../.datasets/tests/mist.pdf'
    temp_path = '../../../.datasets/tests/tmp'
    data = parser.parse_pdf(doc, temp_path=temp_path)
    assert data is not None
    print('Parsed:', doc)



"""
Cannlytics Module | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 7/2/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: This module contains the Cannlytics class,
the entry point into Cannlytics features and functionality.
"""
# Standard imports.
import logging
from os import environ
from typing import Dict, Optional, Union

# External imports
from dotenv import dotenv_values

# Internal imports.
from cannlytics.firebase import initialize_firebase
from cannlytics.metrc import initialize_metrc


class Cannlytics:
    """An instance of this class is the entry point to interface with
    the core Cannlytics logic."""

    def __init__(
            self,
            config: Optional[Union[Dict, str]] = './.env',
            license_number: Optional[str] = None,
            state: Optional[str] = None,
            firebase: Optional[bool] = False,
            metrc: Optional[bool] = False,
            openai: Optional[bool] = False,
    ) -> None:
        """Initialize a Cannlytics class.
        Args:
            config (str, dict): A .env file or configuration with variables:
                `GOOGLE_APPLICATION_CREDENTIALS`
                `LICENSE_NUMBER`
                `METRC_VENDOR_API_KEY`
                `METRC_USER_API_KEY`
                `METRC_STATE`
                `OPENAI_API_KEY`
            license_number (str): A primary license number (optional).
            state (str): A state abbreviation (optional).
            firebase (bool): Initialize the Firebase module, default `False` (optional).
            lims (bool): Initialize the LIMS module, default `False` (optional).
            metrc (bool): Initialize the Metrc module, default `False` (optional).
            paypal (bool): Initialize the PayPal module, default `False` (optional).
            quickbooks (bool): Initialize the QuickBooks module, default `False` (optional).
        """
        if isinstance(config, dict):
            self.config = config
        else:
            self.config = dotenv_values(config)
        self.database = None
        self.license = self.config.get('LICENSE_NUMBER', license_number)
        self.state = self.config.get('METRC_STATE', state)
        self.storage = None
        self.track = None
        self.initialize_logs()
        # TODO: Test / ensure module initialization errors are skipped.
        if firebase:
            self.initialize_firebase(self.config)
        if metrc:
            self.initialize_traceability(
                self.config,
                primary_license=self.license,
                state=self.state
            )
        if openai:
            self.initialize_openai(self.config.get('OPENAI_API_KEY'))


    def create_log(self, message: str):
        """Create a log.
        Args:
            (message): Print a given message to the logs.
        """
        self.logger.debug(message)


    def initialize_logs(self):
        """Initialize logs.
        """
        # Optional: Reduce duplication of logging code (also in the Metrc module)?
        logging.getLogger('cannlytics').handlers.clear()
        logging.basicConfig(
            filename='./tmp/cannlytics.log',
            filemode='w+',
            level=logging.DEBUG,
            format='%(asctime)s %(message)s',
            datefmt='%Y-%m-%dT%H:%M:%S',
        )
        console = logging.StreamHandler()
        console.setLevel(logging.DEBUG)
        self.logger = logging.getLogger('cannlytics')
        self.logger.addHandler(console)


    def initialize_firebase(self, config: Optional[Dict] = None):
        """Initialize a Firebase account for back-end, cloud services.
        Args:
            config (dict): Optional configuration options to override any
                options provided to the Cannlytics class instance.
        Returns:
            (Client): A Firestore database client instance.
        """
        if config is None:
            config = self.config
        credentials = config['GOOGLE_APPLICATION_CREDENTIALS']
        environ['GOOGLE_APPLICATION_CREDENTIALS'] = credentials
        self.database = initialize_firebase()
        self.storage = config.get('FIREBASE_STORAGE_BUCKET')
        # TODO: Map all firebase function to cannlytics.<function_name>, passing DB as an argument.
        self.create_log('Firebase client initialized.')
        return self.database
    

    def initialize_openai(self, openai_api_key=None):
        """Initialize OpenAI with Google Secret Manager or .env variable."""
        import openai
        if openai_api_key is None:
            try:
                import google.auth
                from cannlytics.firebase import access_secret_version
                self.initialize_firebase()
                _, project_id = google.auth.default()
                openai_api_key = access_secret_version(
                    project_id=project_id,
                    secret_id='OPENAI_API_KEY',
                    version_id='latest',
                )
            except:
                openai_api_key = self.config.get('OPENAI_API_KEY')
        openai.api_key = openai_api_key
        self.create_log('OpenAI initialized.')


    def initialize_traceability(self, config=None, primary_license=None, state=None):
        """Initialize the traceability client.
        Args:
            config (dict): The configuration for the traceability client.
            primary_license (str): A primary license to initialize the traceability client.
        """
        if config is None:
            config = self.config
        if state is None:
            state = config.get('METRC_STATE', 'ca')
        self.track = initialize_metrc(
            vendor_api_key=config['METRC_VENDOR_API_KEY'],
            user_api_key=config['METRC_USER_API_KEY'],
            primary_license=primary_license,
            state=state,
        )
        self.create_log('Traceability client initialized.')
        return self.track


    # TODO: Make data and statistics readily available through
    # the main Cannlytics class.



"""
CoADoc | Parse Steep Hill COAs
Copyright (c) 2022 Cannlytics

Authors:
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
    Keegan Skeate <https://github.com/keeganskeate>
Created: 9/23/2022
Updated: 10/10/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse Steep Hill COA PDFs. Validated in:

        ✓ Massachusetts

Data Points:

    ✓ analyses
    ✓ methods
    - {analysis}_status TODO: Record the pass / fail status for each analysis.
    ✓ date_received
    ✓ date_tested
    ✓ lab_results_url
    ✓ metrc_lab_id
    ✓ metrc_source_id
    ✓ producer
    ✓ producer_street
    ✓ producer_city
    ✓ producer_state
    ✓ producer_zipcode
    ✓ producer_license_number
    ✓ product_name
    ✓ product_type
    ✓ results
    ✓ results_hash
    ✓ sample_hash
    ✓ total_cannabinoids
    ✓ total_cbd (Calculated as total_cbd = cbd + 0.877 * cbda)
    ✓ total_thc (Calculated as total_thc = delta_9_thc + 0.877 * thca)
    ✓ total_terpenes

"""
# Standard imports.
from datetime import datetime
import json
import os
import re
from typing import Any, Optional

# External imports.
import pandas as pd
import pdfplumber
import requests

# Internal imports.
from cannlytics import __version__
from cannlytics.data.data import (
    create_hash,
    create_sample_id,
    find_first_value,
)
from cannlytics.data.web import download_google_drive_file
from cannlytics.utils import convert_to_numeric, snake_case
from cannlytics.utils.constants import DECARB
    

# It is assumed that the lab has the following details.
STEEPHILL = {
    'coa_algorithm': 'steephill.py',
    'coa_algorithm_entry_point': 'parse_steephill_coa',
    'url': 'https://www.steephill.com',
    'lims': 'Steep Hill',
    'lab': 'Steep Hill',
    'lab_license_number': 'IL281277',
    'lab_image_url': '',
    'lab_address': '40 Speen Street Suite 301, Framingham, MA, 01701',
    'lab_street': '40 Speen Street Suite 301',
    'lab_city': 'Framingham',
    'lab_county': 'Middlesex',
    'lab_state': 'MA',
    'lab_zipcode': '01701',
    'lab_latitude': 42.312180,
    'lab_longitude': -71.389410,
    'lab_phone': '508-465-3470',
    'lab_email': 'support@steephill.com',
    'lab_website': 'www.steephill.com',
}

# It is assumed that there are the following analyses on each CoA.
STEEPHILL_ANALYSES = {
    'Cannabinoid': {
        'key': 'cannabinoids',
        'columns': ['loq', 'value', 'mg_g', 'mg_serving'],
        'units': 'percent',
    },
    'Heavy Metals': {
        'key': 'heavy_metals',
        'columns': ['loq', 'value', 'limit', 'status'],
        'units': 'ppb',
        # FIXME: Also has limit and status for edibles.
    },
    'Microbial Contaminants': {
        'key': 'microbes',
        'columns': ['value', 'notes', 'tested_at', 'limit', 'status'],
        'units': 'CFU/g',
    },
    'Pathogenic Bacteria': {
        'key': 'microbes',
        'columns': ['value', 'date_tested', 'limit', 'status'],
        'units': 'CFU/g',
    },
    'Mycotoxins': {
        'key': 'mycotoxins',
        'columns': ['loq', 'value', 'limit', 'status'],
        'units': 'ppb',
    },
    'Residual Solvent': {
        'key': 'residual_solvents',
        'columns': ['loq', 'value', 'limit', 'status'],
        'units': 'ppm',
    },
    'Pesticides': {
        'key': 'pesticides',
        'columns': ['loq', 'value', 'limit', 'status'],
        'units': 'ppb',
    },
    'Vitamin E Acetate': {
        'key': 'vitamin_e_acetate',
        'columns': ['loq', 'value', 'limit', 'status'],
        'units': 'ppb',
    },
    'Terpenes': {
        'key': 'terpenes',
        'columns': ['loq', 'value', 'mg_g'],
        'units': 'percent',
    },
}

# It is assumed that the CoA has the following parameters.
STEEPHILL_COA = {
    'fields': {
        'Production Stage': 'product_subtype',
        'Product Class': 'product_type',
        'Ingestion Only': 'cannabinoids_status',
        'Extraction Solvent': 'residual_solvent_status',
        'Retail Name': 'product_name',
    },
}


def parse_steephill_pdf(
        parser,
        doc: Any,
        coa_pdf: Optional[str] = '',
        **kwargs,
    ) -> dict:
    """Parse a Steep Hill CoA PDF.
    Args:
        parser (CoADoc): A CoADoc parsing client.
        doc (str or PDF): A PDF file path or pdfplumber PDF.
        coa_pdf (str): A filename to use for the `coa_pdf` field (optional).
    Returns:
        (dict): The sample data.
    """
    # Read the PDF.
    obs = {}
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = coa_pdf or doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = coa_pdf or report.stream.name.replace('\\', '/').split('/')[-1]

    # Get the lab data.
    front_page = report.pages[0]
    tables = front_page.extract_tables()
    lab_details = tables[0]
    lab_details = [[item for item in x if item] for x in lab_details]
    for row in lab_details:
        for cell in row:
            if 'Submitted:' in cell:
                obs['date_tested'] = cell.split('Submitted:')[-1].strip()
                break

    # Get the client data.
    client_details = tables[1][-1][0].split('\n')
    address_parts = client_details[2].split(',')
    obs['producer'] = client_details[0]
    obs['producer_street'] = client_details[1]
    obs['producer_city'] = address_parts[0]
    obs['producer_state'] = address_parts[-1].strip().split(' ')[0]
    obs['producer_zipcode'] = address_parts[-1].strip().split(' ')[-1]
    obs['producer_license_number'] = client_details[3].split(':')[-1].strip()
    obs['project_id'] = client_details[4].split(':')[-1].strip()
    obs['date_received'] = client_details[5].split(':')[-1].strip()

    # Get the sample IDs.
    sample_ids = tables[2][-1][0].split('\n')
    obs['batch_number'] = sample_ids[0].split(':')[-1].strip()
    obs['metrc_lab_id'] = sample_ids[1].split(':')[-1].strip()
    obs['metrc_source_id'] = sample_ids[2].split(':')[-1].strip()

    # Get the sample data.
    sample_data = tables[3][-1][0].split('\n')
    obs['sample_weight'] = sample_data[0].split(':')[-1].strip()
    obs['serving_size'] = sample_data[1].split(':')[-1].strip()

    # Get the product data, appending long rows to the row before.
    standard_fields = STEEPHILL_COA['fields']
    product_data = tables[4][1:]
    for row in product_data:
        cells = row[0].split('\n')
        for cell in cells:
            parts = cell.split(':')
            try:
                key = standard_fields[parts[0]]
            except KeyError:
                obs[key] = obs[key] + cell
                continue
            value = parts[-1].replace('---', '').strip()
            obs[key] = value

    # Get and standardize the analyses.
    analyses = []
    for table in tables:
        row = table[0]
        if len(row) == 2:
            if row[1] in ['Y', 'P', 'F']:
                analysis = row[0].replace('\n', '')
                analysis = parser.analyses.get(analysis, analysis)
                analyses.append(analysis)

    # Get the methods and results.
    methods, results = [], []
    for page in report.pages[1:]:

        # Stop at the QC section.
        page_text = page.extract_text()
        if 'QA/QC Section' in page_text:
            break

        # Find all of the tables, then extract the lines of each table.
        tables = page.find_tables()
        for table in tables:
            crop = page.within_bbox(table.bbox)
            text = crop.extract_text()
            lines = text.split('\n')
            for line in lines[:-1]: # Skip the footnote.

                # Get the analysis, e.g. "[H] Cannabinoid Profile Metrc...".
                if re.match(r'\[[A-Z]+\] ', line[:4]):
                    analysis_name = line[4:].split('Metrc')[0].strip()
                    analysis_name = analysis_name.replace(' Results', '') \
                        .replace(' Analysis', '') \
                        .replace(' Profile', '')
                    analysis_data = STEEPHILL_ANALYSES[analysis_name]
                    analysis = analysis_data['key']
                    units = analysis_data['units']
                    columns = analysis_data['columns']
                    continue

                # Skip column rows and detail rows.
                elif ';' in line:
                    continue
                elif line.startswith('Cannabinoid'):
                    continue
                elif line.startswith('Terpene'):
                    continue
                elif line.startswith('Analyte'):
                    continue
                elif line.startswith('LOQ'):
                    continue
                elif line.startswith('Datafile'):
                    continue
                elif line.startswith('Analyst'):
                    continue
                elif line.startswith('Result'):
                    continue

                # Collect the methods.
                elif 'were analyzed' in line or 'were measured' in line or \
                    'were incubated' in line or 'was determined' in line:
                    methods.append(line)
                    continue

                # Handle totals appropriately.
                elif line.startswith('Total Terpenes'):
                    value = line.split(' - ')[1]
                    obs['total_terpenes'] = convert_to_numeric(value)
                    continue
                elif line.startswith('Total Available Cannabinoids'):
                    value = value = line.split(' - ')[1]
                    obs['total_cannabinoids'] = convert_to_numeric(value)
                    continue
                elif line.startswith('Total Mycotoxins'):
                    continue

                # Record each result, skipping not-tested analytes.
                elif '(' in line:
                    name = line.split('(')[1].split(')')[0].strip()
                    values = line.split(')')[-1].strip().split(' ')
                    if len(values) < len(columns):
                        continue
                    key = snake_case(name)
                    key = parser.analytes.get(name, parser.analytes.get(key, key))
                    result = {
                        'analysis': analysis,
                        'units': units,
                        'key': key,
                        'name': name,
                    }
                    for i, column in enumerate(columns):
                        result[column] = convert_to_numeric(values[i])
                    results.append(result)
                    continue

                # Handle standard result lines.
                line = line.replace('Not Detected', 'ND')
                first_value = find_first_value(line, [' \d+', 'ND', 'NT'])
                name = line[:first_value].replace('\n', ' ').strip()
                key = snake_case(name)
                key = parser.analytes.get(name, parser.analytes.get(key, key))
                values = line[first_value:].strip().split(' ')
                if len(values) < len(columns):
                    continue
                result = {
                    'analysis': analysis,
                    'units': units,
                    'key': key,
                    'name': name,
                }
                for i, column in enumerate(columns):
                    result[column] = convert_to_numeric(values[i])
                results.append(result)

                # FIXME: microbes status is not correct.

    # Close the report.
    report.close()

    # Calculate total CBD and total THC if present.
    specials = {}
    total_keys = ['cbd', 'cbda', 'delta_9_thc', 'thca']
    for result in results:
        if result['key'] in total_keys:
            specials[result['key']] = result['value']
    if specials.get('cbd') and specials.get('cbda'):
        try:
            acidic = DECARB * specials['cbda']
        except:
            acidic = 0
        try:
            obs['total_cbd'] = specials['cbd'] + acidic
        except:
            obs['total_cbd'] = 'ND'
    if specials.get('delta_9_thc') and specials.get('thca'):
        try:
            acidic = DECARB * specials['thca']
        except:
            acidic = 0
        try:
            obs['total_thc'] = specials['delta_9_thc'] + acidic
        except:
            obs['total_thc'] = 'ND'
        
    # FIXME: Calculate total cannabinoids! Sum of cannabinoids!

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Finish data collection with a freshly minted sample ID.
    obs = {**STEEPHILL, **obs}
    obs['analyses'] = json.dumps(analyses)
    obs['coa_algorithm_version'] = __version__
    obs['coa_parsed_at'] = datetime.now().isoformat()
    obs['methods'] = json.dumps(methods)
    obs['results'] = results
    obs['results_hash'] = create_hash(results)
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=obs.get('producer', obs.get('date_tested', 'cannlytics.eth')),
    )
    obs['sample_hash'] = create_hash(obs)
    return obs


def parse_steephill_url(
        parser,
        url: str,
        headers: Optional[Any] = None,
        temp_path: Optional[str] = '/tmp',
        **kwargs
    ) -> dict:
    """Parse a Steep Hill COA URL.
    Args:
        url (str): The COA URL.
        headers (Any): Optional headers for standardization.
        max_delay (float): The maximum number of seconds to wait
            for the page to load.
        persist (bool): Whether to persist the driver.
            The default is `False`. If you do persist
            the driver, then make sure to call `quit`
            when you are finished.
    Returns:
        (dict): The sample data.
    """
    # Ensure that the URL is direct to Google Drive.
    pdf_url = str(url)
    if not pdf_url.startswith('https://drive.google'):
        response = requests.get(pdf_url, headers=headers)
        pdf_url = response.url
    if not pdf_url.startswith('https://drive.google'):
        raise ValueError('COA URL not recognized as a Steep Hill URL.')

    # Download the PDF from Google Drive.
    temp_pdf = os.path.join(temp_path, 'coa.pdf')
    if not os.path.exists(temp_path): os.makedirs(temp_path)
    drive_id = pdf_url.split('/d/')[-1].split('/')[0]
    download_google_drive_file(drive_id, temp_pdf)

    # Extract the data with the PDF parsing algorithm and record the URL.
    data = parse_steephill_pdf(
        parser,
        temp_pdf,
        coa_pdf=url,
        **kwargs,
    )
    data['lab_results_url'] = pdf_url
    return data


def parse_steephill_coa(
        parser,
        doc: Any,
        **kwargs,
    ) -> dict:
    """Parse a Steep Hill COA PDF or URL.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
        kwargs (arguments): Arguments to pass to the parsing algorithms.
    Returns:
        (dict): The sample data.
    """
    if isinstance(doc, str):
        if doc.startswith('http'):
            return parse_steephill_url(parser, doc, **kwargs)
        elif doc.endswith('.pdf'):
            data = parse_steephill_pdf(parser, doc, **kwargs)
        else:
            data = parse_steephill_pdf(parser, doc, **kwargs)
    else:
        data = parse_steephill_pdf(parser, doc, **kwargs)
    if isinstance(doc, str):
        data['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    elif isinstance(doc, pdfplumber.pdf.PDF):
        data['coa_pdf'] = doc.stream.name.replace('\\', '/').split('/')[-1]
    return data


# === Tests ===
if __name__ == '__main__':

    from cannlytics.data.coas import CoADoc

    # Specify testing constants.
    temp_path = '../../../tests/assets/coas/tmp'
    short_url = 'https://tinyurl.com/mr4cnhm3'
    long_url = 'https://drive.google.com/file/d/10S_odKnB5B76Zhgjd0ALb3ffkSp649sL/view'
    doc = '../../../tests/assets/coas/steep-hill/Dosi-Woah.pdf'
    coas = [short_url, long_url, doc] 

    # # [✓] TEST: Identify Steep Hill COAs.
    # parser = CoADoc()
    # for coa in coas:
    #     lab = parser.identify_lims(coa)
    #     assert lab == 'Steep Hill'

    # # [✓] Parse a Steep Hill COA from a short URL.
    # parser = CoADoc()
    # data = parse_steephill_url(parser, short_url, temp_path=temp_path)
    # assert data is not None
    # print('Parsed:', short_url) 

    # # [✓] Parse a Steep Hill COA from a long URL.
    # parser = CoADoc()
    # data = parse_steephill_url(parser, long_url, temp_path=temp_path)
    # assert data is not None
    # print('Parsed:', long_url) 

    # # [✓] TEST: Parse a Steep Hill COA from a PDF.
    parser = CoADoc()
    doc = '../../../tests/assets/coas/steep-hill/SEL.SL.220601.pdf'
    data = parse_steephill_pdf(parser, doc)
    assert data is not None
    print('Parsed:', doc)

    # # [✓] TEST: Parse each URL and PDF ambiguously.
    # parser = CoADoc()
    # coas = [short_url, long_url, doc]
    # for coa in coas:
    #     data = parser.parse(coa, temp_path=temp_path)
    #     assert data is not None
    # print('✓ Completed Steep Hill COA parsing tests.')



"""
Quality Control Tools | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 2/6/2021  
Updated: 6/23/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Perform various quality control checks and analyses to ensure
that your laboratory is operating as desired.

TODO:
    - Trend analyte results.
    - Create predictions of lab results given available inputs!
    - Statistics for internal standards.
"""


def backup_data():
    """Backup data stored in Firestore."""
    return NotImplementedError


def calculate_relative_percent_diff():
    """Calculate relative perecent difference between two samples."""
    return NotImplementedError


def plot_area_response():
    """Plot area response over time for a group of samples."""
    return NotImplementedError


def plot_deviations():
    """Plot deviations in results for a group of samples."""
    return NotImplementedError


def track_deviations():
    """Track deviations in results for a group of samples."""
    return NotImplementedError


def metrc_reconciliation():
    """Reconcile Metrc data with Firestore data."""
    return NotImplementedError



"""
Data Tools | Cannlytics
Copyright (c) 2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 4/21/2022
Updated: 9/8/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>
"""
# Internal imports.
from hashlib import sha256
import hmac
import json
import os
import re
from typing import Any, Optional

# External imports.
from openpyxl.utils.dataframe import dataframe_to_rows
import pandas as pd

# Internal imports.
from cannlytics.utils import rmerge
from cannlytics.utils.utils import snake_case

# Constants.
DATA_PROVIDER = 'cannlytics'
DATA_URL = 'https://cannlytics.com/api/data'
HASH_ERROR = """Unrecognized `public_key`. Expecting a str, list, dict,
int, bytes, or DataFrame."""

# Library of Cannlytics datasets.
CANNLYTICS_DATASETS = [
    'aggregated-cannabis-test-results',
]

# === Data collection tools. ===

def list_datasets() -> list:
    """Get the Cannlytics Hugging Face data catalogue.
    Returns:
        (list): A list of dataset names.
    """
    return CANNLYTICS_DATASETS


def load_google_sheet(
        sheet_key,
        sheet_name='',
        header_row=1
    ):
    """Load a public Google Sheet by key or URL into a DataFrame. E.g.

        https://docs.google.com/spreadsheets/d/{key}/edit/

    Args:
        sheet_key (str): A Google Sheets URL or the Sheet key found in in the URL.
        sheet_name (str): A name of a particular sheet, the first sheet
            is loaded by default (optional).
        header_row (int): The row of the headers, 1 by default (optional).
    Returns:
        (DataFrame): The Google Sheet data in a DataFrame.    
    Credit: Gianmario Spacagna <https://stackoverflow.com/a/48986530/5021266>
    License: CC BY-SA 3.0 <https://creativecommons.org/licenses/by-sa/3.0/>
    """
    if sheet_key.startswith('https'):
        key = sheet_key.split('/d/')[-1].split('/edit?')[0]
    else:
        key = sheet_key
    url = 'https://docs.google.com/spreadsheets/d/{key}/gviz/tq?tqx=out:csv&sheet={sheet_name}&headers={header_row}'
    url = url.format(
        key=key,
        sheet_name=sheet_name.replace(' ', '%20'),
        header_row=header_row,
    )
    df = pd.read_csv(url)
    return df.drop([col for col in df.columns if col.startswith('Unnamed')], axis=1)



# === Data aggregation tools. ===

def aggregate_datasets(
        directory: str,
        on: Optional[str] = 'sample_id',
        how: Optional[str] = 'left',
        replace: Optional[str] = 'right',
        reverse: Optional[bool] = True,
        concat=False
    ) -> pd.DataFrame:
    """Aggregate datasets. Leverages `rmerge` to combine
    each dataset in the given directory.
    Args:
        directory (string): Required dataset directory.
        on (string): The key to merge datasets, `sample_id` by default (optional).
        how (string): How to merge, `left` by default (optional).
        replace (string): How to replace, `right` by default (optional).
        reverse (bool): Whether to combine in reverse order, `True` by default (optional).
    Returns:
        (DataFrame): The aggregated data.
    """
    all_data = None
    files = os.listdir(directory)
    if reverse:
        files.reverse()
    for filename in files:
        datafile = os.path.join(directory, filename)
        try:
            data = pd.read_excel(datafile)
        except:
            continue
        if all_data is None:
            all_data = data
        else:
            if concat:
                all_data = pd.concat([all_data, data])
            else:
                all_data = rmerge(
                    all_data,
                    data,
                    on=on,
                    how=how,
                    replace=replace,
                )
    return all_data


# === Data cleaning tools. ===

def find_first_value(
        string: str,
        breakpoints: Optional[list]=None,
    ) -> str:
    """Find the first value of a string, be it a digit, a 'ND', '<',
    or other specified breakpoints.
    Args:
        string (str): The string containing a value.
        breakpoints (list): A list of breakpoints (optional).
    Returns:
        (int): Returns the index of the first value.
    """
    if breakpoints is None:
        breakpoints = [' \d+', 'ND', '<']
    detects = []
    for breakpoint in breakpoints:
        try:
            detects.append(string.index(re.search(breakpoint, string).group()))
        except AttributeError:
            pass
    try:
        return min([x for x in detects if x])
    except ValueError:
        return None


def parse_data_block(div, tag='span') -> dict:
    """Parse an HTML data block into a dictionary.
    Args:
        div (bs4.element): An HTML element.
        tag (string): The type of tag that is repeated in the block.
    Returns:
        (dict): A dictionary of key and value pairs.
    """
    data = {}
    for el in div:
        try:
            label = el.find(tag).text
            value = el.text
            value = value.replace(label, '')
            value = value.replace('\n', '').strip()
            label = label.replace(':', '')
            data[snake_case(label)] = value
        except AttributeError:
            pass
    return data


# === Data augmentation tools. ===

def create_hash(
        public_key: Any,
        private_key: Optional[str] = 'cannlytics.eth',
    ) -> str:
    """Create a hash (HMAC-SHA256) that is unique to the provided data.
    Args:
        public_key (str): A string to be used as the public key.
        private_key (str): A string to be used as the private key,
            "cannlytics.eth" by default (optional).
    Returns:
        (str): A sample ID hash.
    References:
        - HMAC: Keyed-Hashing for Message Authentication
        URL: <https://www.ietf.org/rfc/rfc2104.txt>
        - What's the difference between HMAC-SHA256(key, data) and SHA256(key + data)
        URL: <https://security.stackexchange.com/questions/79577/whats-the-difference-between-hmac-sha256key-data-and-sha256key-data>
    """
    if isinstance(public_key, str) or isinstance(public_key, bytes):
        msg = public_key
    elif isinstance(public_key, list):
        msg = str(public_key)
    elif isinstance(public_key, dict):
        msg = json.dumps(public_key)
    elif isinstance(public_key, pd.DataFrame):
        msg = json.dumps(public_key.to_dict('records'))
    elif isinstance(public_key, pd.Series):
        msg = json.dumps(public_key.to_dict())
    elif isinstance(public_key, int) or isinstance(public_key, float):
        msg = str(public_key)
    else:
        raise ValueError(HASH_ERROR)
    try:
        msg = msg.encode()
    except AttributeError:
        pass
    return hmac.new(bytes(private_key, 'UTF-8'), msg, sha256).hexdigest()


def create_sample_id(private_key, public_key, salt='') -> str:
    """Create a hash to be used as a sample ID.
    The standard is to use:
        1. `private_key = results`
        2. `public_key = product_name`
        3. `salt = producer`
    Args:
        private_key (str): A string to be used as the private key.
        public_key (str): A string to be used as the public key.
        salt (str): A string to be used as the salt, '' by default (optional).
    Returns:
        (str): A sample ID hash.
    """
    secret = bytes(private_key, 'UTF-8')
    message = snake_case(public_key) + snake_case(salt)
    sample_id = hmac.new(secret, message.encode(), sha256).hexdigest()
    return sample_id


# === Data analysis tools. ===

def random_sample(data, count=10_000):
    """Create a random sample of a given dataset."""
    # TODO: Implement !
    raise NotImplementedError


def shard_datasets(data, directory, count=10_000):
    """Shard a dataset for ease of use."""
    # TODO: Implement !
    raise NotImplementedError


# === Data saving tools. ===

def write_to_worksheet(ws, values):
    """Write data to an Excel Worksheet.
    Credit: Charlie Clark <https://stackoverflow.com/a/36664027>
    License: CC BY-SA 3.0 <https://creativecommons.org/licenses/by-sa/3.0/>
    Args:
        ws (Worksheet) An openpyxl Worksheet.
        values (list): A list of values to print to the worksheet.
    """
    rows = dataframe_to_rows(pd.DataFrame(values), index=False)
    for r_idx, row in enumerate(rows, 1):
        for c_idx, value in enumerate(row, 1):
            try:
                ws.cell(row=r_idx, column=c_idx, value=value)
            except ValueError:
                ws.cell(row=r_idx, column=c_idx, value=str(value))


if __name__ == '__main__':

    # === Tests ===
    import cannlytics.data

    # [✓] TEST: List all of the available datasets.
    # dataset_names = list_datasets()
    # for name in dataset_names:
    #     assert name in CANNLYTICS_DATASETS

    # [✓] TEST: `aggregate_datasets` with MCR Labs data.
    # data_dir = '../../../.datasets/lab_results/raw_data/mcr_labs'
    # data = aggregate_datasets(data_dir, concat=True)
    # subset = data.loc[~data['results'].isnull()]
    # subset.drop_duplicates(
    #     subset=['sample_id', 'total_cannabinoids'],
    #     keep='last',
    #     inplace=True
    # )
    # timestamp = datetime.now().isoformat()[:19].replace(':', '-')
    # datafile = f'../../../.datasets/lab_results/mcr_labs_test_results-{timestamp}.xlsx'
    # subset.to_excel(datafile, sheet_name='mcr_labs_raw_data')
    # print('Aggregated %i samples.' % len(subset))

    # [✓] TEST: `aggregate_datasets` with PSI Labs data.
    # data_dir = '../../.datasets/lab_results/raw_data/psi_labs'
    # data = aggregate_datasets(data_dir)
    # subset = data.loc[~data['results'].isnull()]
    # subset.drop_duplicates(subset='sample_id', keep='first', inplace=True)
    # subset.to_excel('../../.datasets/lab_results/psi_labs_test_results.xlsx')

    # [✓] TEST: `aggregate_datasets` with SC Labs data.
    # data_dir = '../../.datasets/lab_results/raw_data/sc_labs'
    # data = aggregate_datasets(data_dir, concat=True)
    # subset = data.loc[~data['results'].isnull()]
    # subset.drop_duplicates(subset='sample_id', keep='first', inplace=True)
    # timestamp = datetime.now().isoformat()[:19].replace(':', '-')
    # datafile = f'../../.datasets/lab_results/sc_labs_test_results-{timestamp}.xlsx'
    # subset.to_excel(datafile)
    # print(len(subset))



"""
Metrc Client | Cannlytics
Copyright (c) 2021-2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 1/26/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

This module contains the `Metrc` class responsible for communicating
with the Metrc API.

TODO: Implement the remaining Metrc functionality:

    [ ] GET /transfers/v1/templates/{id}/transporters
    [ ] GET /transfers/v1/templates/{id}/transporters/details
    [ ] Implement any remaining endpoints.
    [ ] Implement return created/updated objects for all endpoints.

"""
# Standard imports.
from datetime import datetime
from json import dumps
import logging
import os
import tempfile

# External imports.
from pandas import read_excel
from requests import Session

# Internal imports.
from .constants import parameters, DEFAULT_HISTORY
from .exceptions import MetrcAPIError
from .models import (
    Delivery,
    Category,
    Employee,
    Facility,
    Item,
    Location,
    Harvest,
    Package,
    Patient,
    Plant,
    PlantBatch,
    LabResult,
    Receipt,
    Strain,
    Transfer,
    TransferTemplate,
    Transaction,
    Waste,
)
from .urls import (
    METRC_API_BASE_URL,
    METRC_API_BASE_URL_TEST,
    METRC_BATCHES_URL,
    METRC_EMPLOYEES_URL,
    METRC_FACILITIES_URL,
    METRC_LOCATIONS_URL,
    METRC_HARVESTS_URL,
    METRC_ITEMS_URL,
    METRC_LAB_RESULTS_URL,
    METRC_PACKAGES_URL,
    METRC_PATIENTS_URL,
    METRC_PLANTS_URL,
    METRC_RECEIPTS_URL,
    METRC_SALES_URL,
    METRC_STRAINS_URL,
    METRC_TRANSACTIONS_URL,
    METRC_TRANSFERS_URL,
    METRC_TRANSFER_PACKAGES_URL,
    METRC_TRANSFER_TEMPLATE_URL,
    METRC_UOM_URL,
)
from ..utils.utils import (
    camel_to_snake,
    clean_dictionary,
    get_timestamp,
)


class Metrc(object):
    """An instance of this class communicates with the Metrc API."""

    def __init__(
            self,
            vendor_api_key,
            user_api_key,
            logs=True,
            primary_license='',
            state='ma',
            test=True,
        ):
        """Initialize a Metrc API client.
        Args:
            vendor_api_key (str): Required Metrc API key, obtained from Metrc
                upon successful certification. The vendor API key is the
                software provider's secret used in every instance, regardless
                of location or licensee.
            user_api_key (str): Required user secret obtained
                from a licensee's Metrc user interface. The user's permissions
                determine the level of access to the Metrc API.
            logs (bool): Whether or not to log Metrc API requests, True by default.
            primary_license (str): A license to use if no license is provided
                on individual requests.
            state (str): The state of the licensee, Oklahoma (ok) by default.
            test (bool): Whether or not to use the test sandbox, True by default.

        Example:

        ```py
        track = metrc.Client(
            vendor_api_key='abc',
            user_api_key='xyz',
            primary_license='123',
            state='ma'
        )
        ```
        """
        self.logs = logs
        self.parameters = parameters
        self.primary_license = primary_license
        self.default_time_period = DEFAULT_HISTORY
        self.state = state
        self.test = test
        self.user_api_key = user_api_key
        self.vendor_api_key = vendor_api_key
        self.session = Session()
        self.session.auth = (vendor_api_key, user_api_key)
        if test:
            self.base = METRC_API_BASE_URL_TEST % state
        else:
            self.base = METRC_API_BASE_URL % state
        if logs:
            self.initialize_logs()


    def request(
            self,
            method,
            endpoint,
            data=None,
            params=None,
        ):
        """Make a request to the Metrc API."""
        url = self.base + endpoint
        try:
            response = getattr(self.session, method)(url, json=data, params=params)
        except ConnectionError:
            self.session = Session()
            self.session.auth = (self.vendor_api_key, self.user_api_key)
            response = getattr(self.session, method)(url, json=data, params=params)
        if self.logs:
            self.create_log(response)
        if response.status_code == 200:
            try:
                return response.json()
            except ValueError:
                return response.text
        else:
            raise MetrcAPIError(response)
    

    def format_params(self, **kwargs):
        """Format Metrc request parameters.
        Returns:
            (dict): Returns the parameters as a dictionary.
        """
        params = {}
        for param in kwargs:
            if kwargs[param]:
                key = self.parameters[param]
                params[key] = kwargs[param]
        return params


    def create_log(self, response):
        """Create a log given an HTTP response.
        Args:
            response (HTTPResponse): An HTTP request response.
        """
        try:
            self.logger.debug(f'Request: {response.request.method} {response.request.url}')
            self.logger.debug(f'Body: {response.request.body}')
            self.logger.debug(f'Status code: {response.status_code}')
            try:
                log = dumps(response.json())
                self.logger.debug(f'Response: {log}')
            except ValueError:
                self.logger.debug(f'Response: {response.text}')
        except KeyError:
            raise MetrcAPIError({'message': '`logs=True` but no logger initialized. Use `client.initialize_logs()`.'})


    def initialize_logs(self):
        """Initialize Metrc logs."""
        timestamp = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
        temp_dir = tempfile.gettempdir()
        temp_file = os.path.join(temp_dir, f'cannlytics-{timestamp}.log')
        logging.getLogger('metrc').handlers.clear()
        logging.basicConfig(
            filename=temp_file,
            filemode='w+',
            level=logging.DEBUG,
            format='%(asctime)s %(message)s',
            datefmt='%Y-%m-%dT%H:%M:%S',
        )
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        self.logger = logging.getLogger('metrc')
        self.logger.addHandler(handler)
        self.logger.debug('Metrc initialized.')


    #-------------------------------------------------------------------
    # Facilities and employees
    #-------------------------------------------------------------------

    def get_facilities(self):
        """Get all facilities."""
        url = METRC_FACILITIES_URL
        response = self.request('get', url)
        return [Facility(self, x) for x in response]


    def get_facility(self, license_number=''):
        """Get a given facility by its license number."""
        url = METRC_FACILITIES_URL
        response = self.request('get', url)
        facility = None
        facilities = [Facility(self, x) for x in response]
        for obs in facilities:
            if obs.license_number == license_number:
                facility = obs
                break
        return facility
    

    def get_employees(self, license_number=''):
        """Get all employees.
        Args:
            license_number (str): A licensee's license number.
        """
        url = METRC_EMPLOYEES_URL
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('get', url, params=params)
        return [Employee(self, x) for x in response]


    #-------------------------------------------------------------------
    # Deliveries
    #-------------------------------------------------------------------

    def create_deliveries(self, data, license_number='', return_obs=False):
        """Create home deliver(ies).
        Args:
            data (list): A list of deliveries (dict) to create.
            license_number (str): A specific license number.
        """
        url = METRC_SALES_URL % 'deliveries'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Return created deliveries.


    def get_deliveries(
            self,
            uid='',
            action='active',
            license_number='',
            start='',
            end='',
            sales_start='',
            sales_end='',
        ):
        """Get sale(s).
        Args:
            uid (str): The UID for a delivery.
            action (str): The action to apply to the delivery, with options:
                `active` or `inactive`.
            license_number (str): A specific license number.
            start (str): An ISO 8601 formatted string to restrict the start
                by the last modified time.
            end (str): An ISO 8601 formatted string to restrict the end
                by the last modified time.
            sales_start (str): An ISO 8601 formatted string to restrict the start
                by the sales time.
            sales_end (str): An ISO 8601 formatted string to restrict the end
                by the sales time.
        Returns:
            (list): Returns a list of Receipts.
        """
        if uid:
            url = METRC_SALES_URL % f'delivery/{uid}'
        else:
            url = METRC_SALES_URL % f'deliveries/{action}'
        params = self.format_params(
            license_number=license_number or self.primary_license,
            start=start,
            end=end,
            sales_start=sales_start,
            sales_end=sales_end,
        )
        response = self.request('get', url, params=params)
        try:
            return Receipt(self, response, license_number)
        except AttributeError:
            return [Receipt(self, x, license_number) for x in response]


    def get_return_reasons(self, license_number=''):
        """Get the possible return reasons for home delivery items.
        Args:
            license_number (str): A specific license number.
        Returns:
            (list): A list of return reasons.
        """
        url = METRC_SALES_URL % 'delivery/returnreasons'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def complete_deliveries(self, data, license_number='', return_obs=False):
        """Complete home delivery(ies).
        Args:
            data (list): A list of deliveries (dict) to complete.
            license_number (str): A specific license number.
        """
        url = METRC_SALES_URL % 'deliveries/complete'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)


    def delete_delivery(self, uid, license_number=''):
        """Delete a home delivery.
        Args:
            uid (str): The UID of a home delivery to delete.
            license_number (str): A specific license number.
        """
        url = METRC_SALES_URL % f'delivery/{uid}'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('delete', url, params=params)


    def update_deliveries(self, data, license_number=''):
        """Update home delivery(ies).
        Args:
            data (list): A list of deliveries (dict) to create.
            license_number (str): A specific license number.
        """
        url = METRC_SALES_URL % 'deliveries'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)

    
    # TODO: Implement state-specific endpoints:
    # - POST /sales/v1/deliveries/retailer/depart
    # - POST /sales/v1/deliveries/retailer/restock
    # - POST /sales/v1/deliveries/retailer/sale
    # - POST /sales/v1/deliveries/retailer/end


    #-------------------------------------------------------------------
    # Harvests
    #-------------------------------------------------------------------

    def get_harvests(
            self,
            uid='',
            action='active',
            license_number='',
            start='',
            end='',
        ):
        """Get harvests.
        Args:
            uid (str): The UID of a harvest, takes precedent over action.
            action (str): A specific filter to apply, including:
                `active`, `onhold`, `inactive`, `waste/types`.
            license_number (str): The licensee's license number.
            start (str): An ISO 8601 formatted string to restrict the start
                by the last modified time.
            end (str): An ISO 8601 formatted string to restrict the end
                by the last modified time.
        """
        if uid:
            url = METRC_HARVESTS_URL % uid
        else:
            url = METRC_HARVESTS_URL % action
        params = self.format_params(license_number=license_number or self.primary_license, start=start, end=end)
        response = self.request('get', url, params=params)
        try:
            return Harvest(self, response, license_number)
        except AttributeError:
            try:
                return [Harvest(self, x, license_number) for x in response]
            except AttributeError:
                return response


    def finish_harvests(self, data, license_number='', return_obs=False):
        """Finish harvests.
        Args:
            data (list): A list of harvests (dict) to finish.
            license_number (str): A specific license number.
        """
        url = METRC_HARVESTS_URL % 'finish'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)


    def unfinish_harvests(self, data, license_number='', return_obs=False):
        """Unfinish harvests.
        Args:
            data (list): A list of harvests (dict) to unfinish.
            license_number (str): A specific license number.
        """
        url = METRC_HARVESTS_URL % 'unfinish'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)


    def remove_waste(self, data, license_number='', return_obs=False):
        """Remove's waste from a harvest.
        Args:
            data (list): A list of waste (dict) to unfinish.
            license_number (str): A specific license number.
        """
        url = METRC_HARVESTS_URL % 'removewaste'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)


    def move_harvests(self, data, license_number='', return_obs=False):
        """Move a harvests.
        Args:
            data (list): A list of harvests (dict) to move.
            license_number (str): A specific license number.
        """
        url = METRC_HARVESTS_URL % 'move'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)


    def create_harvest_packages(self, data, license_number='', return_obs=False):
        """Create packages from a harvest.
        Args:
            data (list): A list of packages (dict) to create.
            license_number (str): A specific license number.
        """
        url = METRC_HARVESTS_URL % 'create/packages'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return packages created.


    def create_harvest_testing_packages(self, data, license_number='', return_obs=False):
        """Create packages from a harvest for testing.
        Args:
            data (list): A list of testing packages (dict) to create.
            license_number (str): A specific license number.
        """
        url = METRC_HARVESTS_URL % 'create/packages/testing'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return packages created.


    #-------------------------------------------------------------------
    # Items
    #-------------------------------------------------------------------

    def get_item_categories(self, license_number=''):
        """Get all item categories.
        Args:
            license_number (str): A specific license number.
        Returns:
            (list): Returns a list of item categories (Category).
        """
        url = METRC_ITEMS_URL % 'categories'
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('get', url, params=params)
        try:
            return [Category(self, x, license_number) for x in response]
        except AttributeError:
            return response


    def get_item(
            self,
            uid='',
            action='active',
            license_number='',
        ):
        """Get an item.
        Args:
            uid (str): The UID of an item.
            action (str): A specific type of item to filter by:
                `active`, `categories`, `brands`.
            license_number (str): A specific license number.
        Returns:
            (Item): Returns a an item.
        """
        response = self.get_items(uid, action=action, license_number=license_number)
        try:
            return response[0]
        except AttributeError:
            return response


    def get_items(
            self,
            uid='',
            action='active',
            license_number='',
        ):
        """Get items.
        Args:
            uid (str): The UID of an item.
            action (str): A specific type of item to filter by:
                `active`, `categories`, `brands`.
            license_number (str): A specific license number.
        Returns:
            (list): Returns a list of items (Item).
        """
        if uid:
            url = METRC_ITEMS_URL % uid
        else:
            url = METRC_ITEMS_URL % action
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('get', url, params=params)
        try:
            return Item(self, response, license_number)
        except AttributeError:
            try:
                return [Item(self, x, license_number) for x in response]
            except AttributeError:
                return response


    def create_item(self, data, license_number='', return_obs=False):
        """Create an item.
        Args:
            data (dict): An item to create.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the newly created item.
        """
        response = self.create_items([data], license_number=license_number)
        if return_obs:
            item_name = data['Name']
            items = self.get_items()
            for item in items:
                if item_name in item.name:
                    return item
            return None
        return response


    def create_items(self, data, license_number='', return_obs=False):
        """Create items.
        Args:
            data (list): A list of items (dict) to create.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the newly created items.
        """
        url = METRC_ITEMS_URL % 'create'
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('post', url, data=data, params=params)
        if not return_obs or not response:
            return response
        else:
            item_names = [x['Name'] for x in data]
            return_items = []
            items = self.get_items()
            for item in items:
                for item_name in item_names:
                    if item_name in item.name:
                        return_items.append(item)
            return return_items


    def update_item(self, data, license_number='', return_obs=False):
        """Update an item.
        Args:
            data (dict): An item to update.
            license_number (str): A specific license number.
        """
        return self.update_items(self, [data], license_number)
        # TODO: Optionally return updated item.


    def update_items(self, data, license_number='', return_obs=False):
        """Update items.
        Args:
            data (list): A list of items (dict) to update.
            license_number (str): A specific license number.
        """
        url = METRC_ITEMS_URL % 'update'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return updated item.


    def delete_item(self, uid, license_number=''):
        """Delete item.
        Args:
            uid (str): The UID of an item to delete.
            license_number (str): A specific license number.
        """
        url = METRC_ITEMS_URL % uid
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('delete', url, params=params)


    # TODO: Implement new item endpoints:
    # - GET /items/v1/brands
    # - GET /items/v1/photo/{id}
    # - POST /items/v1/photo

    #-------------------------------------------------------------------
    # Lab Results
    #-------------------------------------------------------------------

    def get_lab_result(
            self,
            uid='',
            license_number='',
        ):
        """Get lab results.
        Args:
            uid (str): The UID of a lab result.
            license_number (str): A specific license number.
        Returns:
            (LabResult): Returns a a lab result.
        """
        response = self.get_lab_results(uid, license_number=license_number)
        try:
            return response[0]
        except AttributeError:
            return response


    def get_lab_results(
            self,
            uid='',
            license_number='',
        ):
        """Get lab results.
        Args:
            uid (str): The UID for a package.
            license_number (str): A specific license number.
        """
        url = METRC_LAB_RESULTS_URL % 'results'
        params = self.format_params(
            package_id=uid,
            license_number=license_number or self.primary_license,
        )
        response = self.request('get', url, params=params)
        return [LabResult(self, x) for x in response]


    def get_test_types(self, license_number=''):
        """Get required quality assurance analyses.
        Args:
            license_number (str): A specific license number.
        """
        url = METRC_LAB_RESULTS_URL % 'types'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def get_test_statuses(self, license_number=''):
        """Get pre-defined lab statuses.
        Args:
            license_number (str): A specific license number.
        """
        url = METRC_LAB_RESULTS_URL % 'states'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def post_lab_results(self, data, license_number='', return_obs=False):
        """Post lab result(s).
        Args:
            data (list): A list of lab results (dict) to create or update.
            license_number (str): A specific license number.
        """
        url = METRC_LAB_RESULTS_URL % 'record'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return created lab result.
        # track.get_packages(
        #     label=lab_package.label,
        #     license_number=lab.license_number
        # )


    def upload_coas(self, data, license_number='', return_obs=False):
        """Upload lab result CoA(s).
        Args:
            data (list): A list of CoAs (dict) to upload.
            license_number (str): A specific license number.
        """
        url = METRC_LAB_RESULTS_URL % 'labtestdocument'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)


    def release_lab_results(self, data, license_number='', return_obs=False):
        """Release lab result(s).
        Args:
            data (list): A list of package labels (dict) to release lab results.
            license_number (str): A specific license number.
        """
        url = METRC_LAB_RESULTS_URL % 'results/release'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)
    

    def get_coa(
            self,
            uid,
            license_number='',
        ):
        """Get a certificate of analysis for a given test.
        Args:
            uid (str): The UID for a test.
            license_number (str): A specific license number.
        """
        url = METRC_LAB_RESULTS_URL % f'labtestdocument/{uid}'
        params = self.format_params(
            license_number=license_number or self.primary_license,
        )
        return self.request('get', url, params=params)


    #-------------------------------------------------------------------
    # Locations
    #-------------------------------------------------------------------

    def get_location_types(self, license_number=''):
        """Get all location types for a given license.
        Args:
            license_number (str): A specific license number.
        """
        url = METRC_LOCATIONS_URL % 'types'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def get_location(
            self,
            uid='',
            license_number='',
        ):
        """Get a location.
        Args:
            uid (str): The UID of a location.
            license_number (str): A specific license number.
        Returns:
            (Location): Returns a a lab result.
        """
        response = self.get_locations(uid, license_number=license_number)
        try:
            return response[0]
        except AttributeError:
            return response


    def get_locations(
            self,
            uid='',
            action='active',
            license_number='',
        ):
        """Get locations.
        Args:
            uid (str): The UID of a location, takes precedent over action.
            action (str): A specific filter to apply, with options:
                `active`, `types`.
            license_number (str): A specific license number.
        """
        if uid:
            url = METRC_LOCATIONS_URL % uid
        else:
            url = METRC_LOCATIONS_URL % action
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('get', url, params=params)
        try:
            return Location(self, response, license_number)
        except AttributeError:
            try:
                return [Location(self, x, license_number) for x in response]
            except:
                return response


    def create_location(
            self,
            name,
            location_type='default',
            license_number='',
            return_obs=False,
        ):
        """Create location.
        Args:
            name (str): A location name.
            location_type (str): An optional location type:
                `default`, `planting`, or `packing`.
                `default` is assigned by default.
            license_number (str): Optional license number filter.
        """
        return self.create_locations([name], [location_type], license_number)
        # TODO: Optionally return the created location.


    def create_locations(self, names, types=[], license_number='', return_obs=False):
        """Create location(s).
        Args:
            names (list): A list of locations (dict) to create.
            types (list): An optional list of location types:
                `default`, `planting`, or `packing`.
                `default` is assigned by default.
            license_number (str): Optional license number filter.
        """
        data = []
        for index, name in enumerate(names):
            try:
                location_type = types[index]
            except IndexError:
                location_type = 'Default Location Type'
            data.append({
                'Name': name,
                'LocationTypeName': location_type
            })
        url = METRC_LOCATIONS_URL % 'create'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return the created locations.


    def update_locations(self, data, license_number='', return_obs=False):
        """Update location(s).
        Args:
            data (list): A list of locations (dict) to update.
            license_number (str): Optional license number filter.
        """
        url = METRC_LOCATIONS_URL % 'update'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return the updated location.


    def delete_location(self, uid, license_number=''):
        """Delete location.
        Args:
            uid (str): The UID of a location to delete.
            license_number (str): Optional license number filter.
        """
        url = METRC_LOCATIONS_URL % uid
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('delete', url, params=params)


    #-------------------------------------------------------------------
    # Packages
    #-------------------------------------------------------------------

    def get_adjustment_reasons(self, license_number=''):
        """Get reasons for adjusting packages."""
        objs = self.get_packages(
            action='adjust/reasons',
            license_number=license_number,
        )
        try:
            return [obj.to_dict() for obj in objs]
        except AttributeError:
            return objs


    def get_package_types(self, license_number=''):
        """Get all facilities.
        Args:
            license_number (str): A specific license number.
        """
        url = METRC_PACKAGES_URL % 'types'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def get_package(
            self,
            uid='',
            label='active',
            action='active',
            license_number='',
        ):
        """Get a package.
        Args:
            uid (str): The UID of an item.
            label (str): The tag label for a package.
            action (str): A specific type of item to filter by:
                `active`, `categories`, `brands`.
            license_number (str): A specific license number.
        Returns:
            (Item): Returns an item (Item).
        """
        response = self.get_packages(uid, label=label, action=action, license_number=license_number)
        try:
            return response[0]
        except AttributeError:
            return response


    def get_packages(
            self,
            uid='',
            label='',
            action='active',
            license_number='',
            start='',
            end='',
        ):
        """Get package(s).
        Args:
            uid (str): The UID for a package.
            label (str): The tag label for a package.
            license_number (str): A specific license number.
            action (str): `active`, `onhold`, `inactive`, `types`,
                `adjust/reasons`.
            start (str): Optional ISO date to restrict earliest modified transfers.
            end (str): Optional ISO date to restrict latest modified transfers.
        Returns:
            (list): Returns a list of packages (Packages).
        """
        if uid:
            url = METRC_PACKAGES_URL % uid
        elif label:
            url = METRC_PACKAGES_URL % label
        else:
            url = METRC_PACKAGES_URL % action
        license_number = license_number or self.primary_license
        params = self.format_params(license_number=license_number, start=start, end=end)
        response = self.request('get', url, params=params)
        try:
            return Package(self, response, license_number)
        except AttributeError:
            try:
                return [Package(self, x, license_number) for x in response]
            except AttributeError:
                return response


    def create_package(
            self,
            data,
            license_number='',
            testing=False,
            plantings=False,
            return_obs=False,
        ):
        """Create a single package."""
        return self.create_packages(
            [data],
            license_number=license_number,
            testing=testing,
            plantings=plantings,
            return_obs=return_obs,
        )
        

    def create_packages(
            self,
            data,
            license_number='',
            testing=False,
            plantings=False,
            return_obs=False,
        ):
        """Create packages.
        Args:
            data (list): A list of packages (dict) to create.
            license_number (str): A specific license number.
            testing (bool): If the packages are for QA testing.
            plantings (bool): If the packages are for planting.
        """
        url = METRC_PACKAGES_URL % 'create'
        if testing:
            url += '/testing'
        elif plantings:
            url += '/plantings'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return created packages.


    def update_package(self, data, license_number='', return_obs=False):
        """Update a given package."""
        return self.update_packages([data], license_number=license_number, return_obs=return_obs)


    def update_packages(self, data, license_number='', return_obs=False):
        """Update packages.
        Args:
            data (list): A list of packages (dict) to update.
            license_number (str): A specific license number.
        """
        url = METRC_PACKAGES_URL % 'update'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return updated packages.


    def change_package_items(self, data, license_number='', return_obs=False):
        """Update package items.
        Args:
            data (list): A list of package items (dict) to update.
            license_number (str): A specific license number.
        """
        url = METRC_PACKAGES_URL % 'change/item'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return updated packages.


    def change_package_locations(self, data, license_number='', return_obs=False):
        """Update package item location(s).
        Args:
            data (list): A list of package items (dict) to move.
            license_number (str): A specific license number.
        """
        url = METRC_PACKAGES_URL % 'change/locations'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return updated packages.


    def manage_packages(self, data, action='adjust', license_number='', return_obs=False):
        """Adjust package(s).
        Args:
            data (list): A list of packages (dict) to manage.
            license_number (str): A specific license number.
            action (str): The action to apply to the packages, with options:
                `adjust`, `finish`, `unfinish`, `remediate`.
        """
        url = METRC_PACKAGES_URL % action
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return updated packages.


    def update_package_notes(self, data, license_number='', return_obs=False):
        """Update package note(s).
        Args:
            data (list): A list of package notes (dict) to update.
            license_number (str): A specific license number.
        """
        url = METRC_PACKAGES_URL % 'change/note'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)
        # TODO: Optionally return updated packages.


    def create_plant_batches_from_packages(self, data, license_number='', return_obs=False):
        """Create plant batch(es) from given package(s).
        Args:
            data (list): A list of plant batches (dict) to create.
            license_number (str): A specific license number.
        """
        url = METRC_PACKAGES_URL % 'create/plantings'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return updated packages.


    #-------------------------------------------------------------------
    # Patients
    #-------------------------------------------------------------------

    def get_patient(self, uid, license_number=''):
        return self.get_patients(uid, license_number=license_number)


    def get_patients(self, uid='', action='active', license_number=''):
        """Get licensee member patients.
        Args:
            uid (str): A UID for a patient.
            action (str): An optional filter to apply: `active` or
                `statuses/{patientLicenseNumber}` in certain states (e.g. mi).
            license_number (str): A licensee's license number to filter by.
        """
        if uid:
            url = METRC_PATIENTS_URL % uid
        else:
            url = METRC_PATIENTS_URL % action
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('get', url, params=params)
        try:
            return Patient(self, response, license_number)
        except AttributeError:
            try:
                return [Patient(self, x, license_number) for x in response]
            except AttributeError:
                return response


    def create_patient(self, data, license_number='', return_obs=False):
        """Create a given patient."""
        return self.create_patients([data], license_number, return_obs)


    def create_patients(self, data, license_number='', return_obs=False):
        """Create patient(s).
        Args:
            data (list): A list of patient (dict) to add.
        """
        url = METRC_PATIENTS_URL % 'add'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return created patient.


    def update_patients(self, data, license_number='', return_obs=False):
        """Update strain(s).
        Args:
            data (list): A list of patients (dict) to update.
        """
        url = METRC_PATIENTS_URL % 'update'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)
        # TODO: Optionally return updated patients.


    def delete_patient(self, uid, license_number=''):
        """Delete patient.
        Args:
            uid (str): The UID of a patient to delete.
        """
        url = METRC_PATIENTS_URL % uid
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('delete', url, params=params)
    

    def get_patient_registration_locations(self):
        """Get patient registration locations."""
        url = METRC_SALES_URL % f'patientregistration/locations'
        return self.request('get', url)


    #-------------------------------------------------------------------
    # Plant Batches
    #-------------------------------------------------------------------

    def create_plant_batch(self, data, license_number='', return_obs=False):
        """Create a plant batch.
        Args:
            data (dict): A plant batch to create.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created
                plant batch.
        Returns:
            (PlantBatch): Returns a plant batch class.
        """
        response = self.create_plant_batches([data], license_number=license_number)
        if return_obs:
            name = data['Name']
            start = get_timestamp(past=self.default_time_period, zone=self.state)
            objects = self.get_batches(start=start, license_number=license_number)
            for obs in objects:
                if obs.name == name:
                    return obs
            return None
        return response


    def create_plant_batches(self, data, license_number='', return_obs=False):
        """Create plant batches.
        Args:
            data (list): A list of plant batches (dict) to create.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created
                plant batches.
        Returns:
            (list): Returns a list of plant batch (PlantBatch) classes.
        """
        if self.state == 'ca':
            raise MetrcAPIError({'message': 'The request POST /plantbatches/v1/createplantings will not work in California due to "CanCreateOpeningBalancePlantBatches": false, this request is used in other states that allow Plant Batch creation.'})
        response = self.manage_batches(data, 'createplantings', license_number=license_number)
        if not return_obs:
            return response
        else:
            names = [x['Name'] for x in data]
            return_obs = []
            start = get_timestamp(past=self.default_time_period, zone=self.state)
            objects = self.get_batches(start=start, license_number=license_number)
            for obs in objects:
                for name in names:
                    if obs.name == name:
                        return_obs.append(obs)
            return return_obs


    def get_batch_types(self, license_number=''):
        """Get plant batch types.
        Args:
            license_number (str): A specific license number.
        """
        return self.get_batches(action='types', license_number=license_number)


    def get_batches(
            self,
            uid='',
            action='active',
            license_number='',
            start='',
            end=''
        ):
        """Get plant batches(s).
        Args:
            uid (str): The UID for a plant batch.
            action (str): The action to apply to the plants, with options:
                `active`, `inactive`, `types`.
            license_number (str): A specific license number.
            start (str): An ISO 8601 formatted string to restrict the start
                by the last modified time.
            end (str): An ISO 8601 formatted string to restrict the end
                by the last modified time.
        """
        if uid:
            url = METRC_BATCHES_URL % uid
        else:
            url = METRC_BATCHES_URL % action
        params = self.format_params(license_number=license_number or self.primary_license, start=start, end=end)
        response = self.request('get', url, params=params)
        try:
            return PlantBatch(self, response, license_number)
        except AttributeError:
            try:
                return [PlantBatch(self, x, license_number) for x in response]
            except AttributeError:
                return response


    def manage_batches(self, data, action, license_number='', from_mother=False, return_obs=False):
        """Manage plant batch(es) by applying a given action.
        Args:
            data (list): A list of plants (dict) to manage.
            action (str): The action to apply to the plants, with options:
                `createplantings`, `createpackages`, `split`,
                `create/packages/frommotherplant`, `changegrowthphase`,
                `additives`, `destroy`.
            from_mother (bool): An optional parameter for the
                `createpackages` action.
        """
        url = METRC_BATCHES_URL % action
        params = self.format_params(from_mother=from_mother, license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return updated or created objects.


    def add_batch_additives(self, data, license_number='', return_obs=False):
        """Add additives to a given batch."""
        if isinstance(data, dict):
            objs = [data]
        else:
            objs = data
        return self.manage_batches(objs, 'additives', license_number or self.primary_license)


    def change_batch_growth_phase(self, data, license_number='', return_obs=False):
        """Change the growth phase of given batch(es)."""
        if isinstance(data, dict):
            objs = [data]
        else:
            objs = data
        return self.manage_batches(objs, 'changegrowthphase', license_number or self.primary_license)


    def create_plantings(self, data, license_number='', return_obs=False):
        """Create plantings from given batch."""
        if isinstance(data, dict):
            objs = [data]
        else:
            objs = data
        return self.manage_batches(objs, 'createplantings', license_number or self.primary_license)


    def create_plant_package_from_batch(self, data, license_number='', from_mother_plant=False, return_obs=False):
        """Create a plant package from a batch.
        Args:
            data (dict): The plant package data.
            license_number (str): A specific license number.
        """
        if from_mother_plant:
            return self.manage_batches([data], 'create/packages/frommotherplant', license_number=license_number)
        else:
            return self.manage_batches([data], 'createpackages', license_number=license_number)        
        # TODO: Optionally return created package.
   

    def destroy_batch_plants(self, data, license_number='', return_obs=False):
        """Destroy plants in a given batch."""
        if isinstance(data, dict):
            objs = [data]
        else:
            objs = data
        return self.manage_batches(objs, 'destroy', license_number or self.primary_license)


    def move_batches(self, data, license_number='', return_obs=False):
        """Move plant batch(es).
        Args:
            data (list): A list of plant batches (dict) to move.
        """
        url = METRC_BATCHES_URL % 'moveplantbatches'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)
        # TODO: Optionally return moved plant batch.


    def split_batch(self, data, license_number='', return_obs=False):
        """Split a given batch.
        Args:
            data (dict): Batch split data.
            license_number (str): A specific license number.
        """
        return self.split_batches([data], license_number=license_number)
        # TODO: Optionally return new batch.


    def split_batches(self, data, license_number='', return_obs=False):
        """Split multiple batches.
        Args:
            data (list): A list of batch splits (dict).
            license_number (str): A specific license number.
        """
        return self.manage_batches(data, action='split', license_number=license_number)
        # TODO: Optionally return new batches.


    #-------------------------------------------------------------------
    # Plants
    #-------------------------------------------------------------------

    def create_plant(self, data, license_number='', return_obs=False):
        """Create a plant.
        Args:
            data (dict): A plant to create.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created
                plant.
        Returns:
            (Plant): Returns a plant batch class.
        """
        response = self.create_plants([data], license_number=license_number)
        if return_obs:
            label = data['PlantLabel']
            return self.get_plants(label=label, license_number=license_number)
        return response


    def create_plants(self, data, license_number='', return_obs=False):
        """Use a plant to create an immature plant batch.
        Args:
            data (list): A list of plants (dict) to create.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created
                plants.
        Returns:
            (list): Returns a list of plants (Plants) classes.
        """
        url = METRC_PLANTS_URL % 'create/plantings'
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('post', url, data=data, params=params)
        if not return_obs:
            return response
        else:
            names = [x['PlantLabel'] for x in data]
            return_obs = []
            start = get_timestamp(past=self.default_time_period, zone=self.state)
            objects = self.get_plants(
                action='vegetative',
                start=start,
                license_number=license_number,
            )
            for obs in objects:
                for name in names:
                    if obs.label == name:
                        return_obs.append(obs)
            return return_obs


    def create_plant_package(self, data, license_number='', return_obs=False):
        """Create plant package.
        Args:
            data (list): A list of plant packages (dict) to create.
            license_number (str): A specific license number.
        """
        return self.create_plant_packages([data], license_number)


    def create_plant_packages(self, data, license_number='', return_obs=False):
        """Create plant packages.
        Args:
            data (list): A list of plant packages (dict) to create.
            license_number (str): A specific license number.
        """
        url = METRC_PLANTS_URL % 'create/plantbatch/packages'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)


    def get_plants(
            self,
            uid='',
            label='',
            action='',
            license_number='',
            start='',
            end=''
        ):
        """Get plant(s).
        Args:
            uid (str): The UID for a plant.
            label (str): The label for a given plant.
            action (str): A specific filter to apply, with options:
                `vegetative`, `flowering`, `onhold`,
                `inactive`, `additives`, `additives/types`,
                `growthphases`, `waste/methods`, `waste/reasons`.
            license_number (str): A specific license number.
            start (str): An ISO 8601 formatted string to restrict the start
                by the last modified time.
            end (str): An ISO 8601 formatted string to restrict the end
                by the last modified time.
        """
        if uid:
            url = METRC_PLANTS_URL % uid
        elif label:
            url = METRC_PLANTS_URL % label
        else:
            url = METRC_PLANTS_URL % action
        params = self.format_params(license_number=license_number or self.primary_license, start=start, end=end)
        response = self.request('get', url, params=params)
        try:
            return Plant(self, response, license_number)
        except AttributeError:
            try:
                return [Plant(self, x, license_number) for x in response]
            except AttributeError:
                return response


    def manage_plants(self, data, action, license_number='', return_obs=False):
        """Manage plant(s) by applying a given action.
        Args:
            data (list): A list of plants (dict) to manage.
            action (str): The action to apply to the plants, with options:
                `moveplants`, `changegrowthphases`, `destroyplants`,
                `additives`, `additives/bylocation`,
                `create/plantings`, `create/plantbatch/packages`,
                `manicureplants`, `harvestplants`.
            license_number (str): A specific license number.
        """
        url = METRC_PLANTS_URL % action
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return updated plants.


    def move_plants(self, data, license_number='', return_obs=False):
        """Move multiple plants.
        Args:
            data (list): A list of plant move data (dict).
            license_number (str): A specific license number.
        """
        return self.manage_plants(data, action='moveplants', license_number=license_number)
        # TODO: Optionally return updated plants.


    def destroy_plants(self, data, license_number=''):
        """Destroy plants."""
        return self.manage_plants(data, action='destroyplants', license_number=license_number)


    def flower_plants(self, data, license_number='', return_obs=False):
        """Flower plants."""
        return self.manage_plants(data, action='changegrowthphases', license_number=license_number)


    def harvest_plants(self, data, license_number='', return_obs=False):
        """Harvest plants."""
        return self.manage_plants(data, action='harvestplants', license_number=license_number)


    def manicure_plants(self, data, license_number='', return_obs=False):
        """Manicure plants."""
        return self.manage_plants(data, action='manicureplants', license_number=license_number)


    def add_plant_additives(self, data, license_number='', return_obs=False):
        """Add additive(s) to given plant(s)."""
        return self.manage_plants(data, action='additives', license_number=license_number)


    def get_additive_types(self, license_number='', return_obs=False):
        """Get additive types."""
        return self.get_plants(action='additives/types', license_number=license_number)
    

    def get_growth_phases(self, license_number='', return_obs=False):
        """Get growth phases."""
        return self.get_plants(action='growthphases', license_number=license_number)
    


    # TODO: Implement endpoint: "additives/bylocation"


    #-------------------------------------------------------------------
    # Sales
    #-------------------------------------------------------------------

    def get_receipts(
            self,
            uid='',
            action='active',
            license_number='',
            start='',
            end='',
            sales_start='',
            sales_end='',
        ):
        """Get sale(s).
        Args:
            uid (str): The UID for a plant batch.
            action (str): The action to apply to the plants, with options:
                `active` or `inactive`
            license_number (str): A specific license number.
            start (str): An ISO 8601 formatted string to restrict the start
                by the last modified time.
            end (str): An ISO 8601 formatted string to restrict the end
                by the last modified time.
            sales_start (str): An ISO 8601 formatted string to restrict the start
                by the sales time.
            sales_end (str): An ISO 8601 formatted string to restrict the end
                by the sales time.
        Returns:
            (list): Returns a list of Receipts or a singular Receipt.
        """
        if uid:
            url = METRC_RECEIPTS_URL % uid
        else:
            url = METRC_RECEIPTS_URL % action
        params = self.format_params(
            license_number=license_number or self.primary_license,
            start=start,
            end=end,
            sales_start=sales_start,
            sales_end=sales_end,
        )
        response = self.request('get', url, params=params)
        try:
            return Receipt(self, response, license_number)
        except AttributeError:
            return [Receipt(self, x, license_number) for x in response]


    def get_transactions(
            self,
            license_number='',
            start='',
            end='',
        ):
        """Get transaction(s).
        Args:
            license_number (str): A specific license number.
            start (str): An ISO 8601 formatted string to restrict the start
                by the sales time.
            end (str): An ISO 8601 formatted string to restrict the end
                by the sales time.
        Returns:
            (list): Returns either a list of Transactions or a singular Transaction.
        """
        if start and end:
            url = METRC_TRANSACTIONS_URL % f'{start}/{end}'
        else:
            url = METRC_SALES_URL % 'transactions'
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('get', url, params=params)
        try:
            return Transaction(self, response, license_number)
        except AttributeError:
            return [Transaction(self, x, license_number) for x in response]


    def get_customer_types(self, license_number=''):
        """Get all customer types.
        Args:
            license_number (str): A specific license number.
        Returns:
            (list): Returns a list of customer types (dict).
        """
        url = METRC_SALES_URL % 'customertypes'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def create_receipt(self, data, license_number='', return_obs=False):
        """Create a receipt.
        Args:
            data (dict): A receipts (dict) to create.
            license_number (str): A specific license number.
        """
        return self.create_receipts(data, license_number, return_obs)


    def create_receipts(self, data, license_number='', return_obs=False):
        """Create receipt(s).
        Args:
            data (list): A list of receipts (dict) to create.
            license_number (str): A specific license number.
        """
        url = METRC_SALES_URL % 'receipts'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return the created receipts.


    def update_receipts(self, data, license_number='', return_obs=False):
        """Update receipt(s).
        Args:
            data (list): A list of receipts (dict) to update.
            license_number (str): A specific license number.
        """
        url = METRC_SALES_URL % 'receipts'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)
        # TODO: Optionally return the updated receipts.


    def delete_receipt(self, uid, license_number=''):
        """Delete receipt.
        Args:
            uid (str): The UID of a receipt to delete.
            license_number (str): A specific license number.
        """
        url = METRC_SALES_URL % f'receipts/{uid}'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('delete', url, params=params)


    def create_transactions(self, data, date=None, license_number='', return_obs=False):
        """Create transaction(s).
        Args:
            data (list): A list of transactions (dict) to create.
            date (str): An ISO 8601 formatted string of the transaction date.
            license_number (str): A specific license number.
        Return:
            (Transaction): Return the created transaction if `return_obs=True`.
        """
        if date is None:
            date = get_timestamp(zone=self.state)
        url = METRC_TRANSACTIONS_URL % date
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return the created transactions.


    def update_transactions(self, data, date=None, license_number='', return_obs=False):
        """Update transaction(s).
        Args:
            data (list): A list of transactions (dict) to update.
            date (str): An ISO 8601 formatted string of the transaction date.
            license_number (str): A specific license number.
        Return:
            (list): Return a list of transactions (Transaction) if `return_obs=True`.
        """
        if date is None:
            date = get_timestamp(zone=self.state)
        url = METRC_TRANSACTIONS_URL % date
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)
        # TODO: Optionally return the updated transactions.


    #-------------------------------------------------------------------
    # Strains
    #-------------------------------------------------------------------

    def get_strains(self, uid='', action='active', license_number=''):
        """Get strains.
        Args:
            uid (str): A UID for a strain.
            action (str): An optional filter to apply: `active`.
            license_number (str): A licensee's license number to filter by.
        """
        if uid:
            url = METRC_STRAINS_URL % uid
        else:
            url = METRC_STRAINS_URL % action
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('get', url, params=params)
        try:
            return Strain(self, response, license_number)
        except AttributeError:
            return [Strain(self, x, license_number) for x in response]


    def create_strain(self, data, license_number='', return_obs=False):
        """Create a strain.
        Args:
            data (dict): A strain to create.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created
                strain (Strain).
        Returns:
            (Strain): Returns a strain class.
        """
        response = self.create_strains(
            [data],
            license_number=license_number,
            return_obs=return_obs,
        )
        if response:
            return response[0]
        return response


    def create_strains(self, data, license_number='', return_obs=False):
        """Create strain(s).
        Args:
            data (list): A list of strains (dict) to create.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created
                strains (Strain).
        Returns:
            (list): Returns a list of strains (Strains) classes.
        """
        url = METRC_STRAINS_URL % 'create'
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('post', url, data=data, params=params)
        if not return_obs:
            return response
        else:
            names = [x['Name'] for x in data]
            return_obs = []
            objects = self.get_strains(license_number=license_number)
            for obs in objects:
                for name in names:
                    if obs.name == name:
                        return_obs.append(obs)
            return return_obs


    def update_strain(self, data, license_number='', return_obs=False):
        """Update strain.
        Args:
            data (list): A strain (dict) to update.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created strains.
        Returns:
            (Strain): Returns the updated strain.
        """
        response = self.update_strains(
            [data],
            license_number=license_number,
            return_obs=return_obs,
        )
        if response:
            return response[0]
        return response


    def update_strains(self, data, license_number='', return_obs=False):
        """Update strain(s).
        Args:
            data (list): A list of strains (dict) to update.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created strains.
        Returns:
            (list): Returns a list of strains (Strain) classes.
        """
        url = METRC_STRAINS_URL % 'update'
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('post', url, data=data, params=params)
        if not return_obs:
            return response
        else:
            ids = [x['Id'] for x in data]
            return_obs = []
            for obs_id in ids:
                obs = self.get_strains(uid=obs_id, license_number=license_number)
                return_obs.append(obs)
            return return_obs


    def delete_strain(self, uid, license_number=''):
        """Delete strain.
        Args:
            uid (str): The UID of a strain to delete.
            license_number (str): A specific license number.
        """
        url = METRC_STRAINS_URL % uid
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('delete', url, params=params)


    #-------------------------------------------------------------------
    # Transfers
    #-------------------------------------------------------------------

    def get_transfers(
            self,
            uid='',
            transfer_type='incoming',
            license_number='',
            start='',
            end='',
        ):
        """Get transfers.
        Args:
            uid (str): The UID for a transfer, takes precedent in query.
            transfer_type (str): The type of transfer:
                `incoming`, `outgoing`, or `rejected`.
            license_number (str): A specific license number.
            start (str): Optional ISO date to restrict earliest modified transfers.
            end (str): Optional ISO date to restrict latest modified transfers.
        """
        if uid:
            url = METRC_TRANSFERS_URL % f'{uid}/deliveries'
        else:
            url = METRC_TRANSFERS_URL % transfer_type
        params = self.format_params(license_number=license_number or self.primary_license, start=start, end=end)
        response = self.request('get', url, params=params)
        try:
            return Transfer(self, response, license_number)
        except AttributeError:
            return [Transfer(self, x, license_number) for x in response]


    def get_transfer_packages(self, uid, license_number='', action='packages'):
        """Get shipments.
        Args:
            uid (str): Required UID of a shipment.
            license_number (str): A specific license number.
            action (str): The filter to apply to transfers:
                `packages`, `packages/wholesale`, `requiredlabtestbatches`.
        """
        if action == 'requiredlabtestbatches':
            url = METRC_TRANSFER_PACKAGES_URL % (f'package/{uid}', action)
        else:
            url = METRC_TRANSFER_PACKAGES_URL % (uid, action)
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def get_transfer_types(self, license_number=''):
        """Get all transfer types.
        Args:
            license_number (str): A specific license number.
        Returns:
            (list): Returns transfer types (dict).
        """
        url = METRC_TRANSFERS_URL % 'types'
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('get', url, params=params)
        try:
            return [clean_dictionary(x, camel_to_snake) for x in response]
        except:
            return response


    def get_package_statuses(self, license_number=''):
        """Get all package status choices.
        Args:
            license_number (str): A specific license number.
        Returns:
            (list): Returns package statuses (dict).
        """
        url = METRC_TRANSFERS_URL % 'delivery/packages/states'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def get_transporters(self, uid):
        """Get the data for a transporter.
        Args:
            uid (str): The ID of the shipment delivery.
        """
        url = METRC_TRANSFERS_URL % f'{uid}/transporters'
        return self.request('get', url)


    def get_transporter_details(self, uid):
        """Get the details of the transporter driver and vehicle.
        Args:
            uid (str): The ID of the shipment delivery.
        """
        url = METRC_TRANSFERS_URL % f'{uid}/transporters/details'
        return self.request('get', url)


    def create_transfer(self, data, license_number='', return_obs=False):
        """Create a transfer.
        Args:
            data (dict): A transfer to create.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created
                transfer.
        Returns:
            (Transfer): Returns the created transfer.
        """
        response = self.create_transfers(
            [data],
            license_number=license_number,
            return_obs=return_obs,
        )
        if response:
            return response[0]
        return response


    def create_transfers(self, data, license_number='', return_obs=False):
        """Create transfer(s).
        Args:
            data (list): A list of transfers (dict) to create.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created
                transfer.
        Returns:
            (list): Returns a list of transfers (Transfer).
        """
        url = METRC_TRANSFERS_URL % 'external/incoming'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return the created transfers.


    def update_transfer(self, data, license_number='', return_obs=False):
        """Update a given transfer.
        Args:
            data (dict): A transfer to update.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the created transfer.
        Returns:
            (Transfer): Returns transfer class.
        """
        response = self.update_transfers(
            [data],
            license_number=license_number,
            return_obs=return_obs,
        )
        if response:
            return response[0]
        return response


    def update_transfers(self, data, license_number='', return_obs=False):
        """Update transfer(s).
        Args:
            data (list): A list of transfers (dict) to update.
            license_number (str): A specific license number.
            return_obs (bool): Whether or not to get and return the updated transfers.
        Returns:
            (list): Returns a list of transfers (Transfer).
        """
        url = METRC_TRANSFERS_URL % 'external/incoming'
        params = self.format_params(license_number=license_number or self.primary_license)
        response = self.request('put', url, data=data, params=params)
        if not return_obs:
            return response
        else:
            ids = [x['TransferId'] for x in data]
            return_obs = []
            for obs_id in ids:
                obs = self.get_transfers(uid=obs_id, license_number=license_number)
                return_obs.append(obs)
            return return_obs


    def delete_transfer(self, uid, license_number=''):
        """Delete transfer.
        Args:
            uid (str): The UID of a transfer to delete.
            license_number (str): A specific license number.
        """
        url = METRC_TRANSFERS_URL % f'external/incoming/{uid}'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('delete', url, params=params)


    #-------------------------------------------------------------------
    # Transfer Templates
    #-------------------------------------------------------------------

    # TODO: Implement remaining transfer template endpoints:
    # - GET /transfers/v1/templates/{id}/transporters
    # - GET /transfers/v1/templates/{id}/transporters/details

    def get_transfer_templates(
            self,
            uid='',
            action='',
            license_number='',
            start='',
            end='',
        ):
        """Get transfer template(s).
        Args:
            uid (str): A UID for a transfer template.
            action (str): An optional filter to apply: `deliveries`, `packages`.
            license_number (str): A specific license number.
            start (str): An ISO 8601 formatted string to restrict the start
                by the sales time.
            end (str): An ISO 8601 formatted string to restrict the end
                by the sales time.
        """
        if action == 'deliveries':
            url = METRC_TRANSFER_TEMPLATE_URL % f'{uid}/deliveries'
        elif action == 'packages':
            url = METRC_TRANSFER_TEMPLATE_URL % f'delivery/{uid}/packages'
        else:
            url = (METRC_TRANSFER_TEMPLATE_URL % '').rstrip('/')
        params = self.format_params(license_number=license_number or self.primary_license, start=start, end=end)
        response = self.request('get', url, params=params)
        try:
            return TransferTemplate(self, response, license_number)
        except AttributeError:
            return [TransferTemplate(self, x, license_number) for x in response]


    def create_transfer_templates(self, data, license_number='', return_obs=False):
        """Create transfer_template(s).
        Args:
            data (list): A list of transfer templates (dict) to create.
        """
        url = (METRC_TRANSFER_TEMPLATE_URL % '').rstrip('/')
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('post', url, data=data, params=params)
        # TODO: Optionally return the created transfer templates.


    def update_transfer_templates(self, data, license_number='', return_obs=False):
        """Update transfer template(s).
        Args:
            data (list): A list of transfer templates (dict) to update.
        """
        url = (METRC_TRANSFER_TEMPLATE_URL % '').rstrip('/')
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('put', url, data=data, params=params)
        # TODO: Optionally return the updated transfer templates.


    def delete_transfer_template(self, uid, license_number=''):
        """Delete transfer template.
        Args:
            uid (str): The UID of a transfer template to delete.
        """
        url = METRC_TRANSFER_TEMPLATE_URL % uid
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('delete', url, params=params)


    #-------------------------------------------------------------------
    # Waste
    #-------------------------------------------------------------------

    def get_waste_methods(self, license_number=''):
        """Get all waste methods for a given license.
        Args:
            license_number (str): A specific license number.
        """
        url = METRC_PLANTS_URL % 'waste/methods'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def get_waste_reasons(self, license_number=''):
        """Get all waste reasons for plants for a given license.
        Args:
            license_number (str): A specific license number.
        """
        url = METRC_PLANTS_URL % 'waste/reasons'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def get_waste_types(self, license_number=''):
        """Get all waste types for harvests for a given license.
        Args:
            license_number (str): A specific license number.
        """
        url = METRC_HARVESTS_URL % 'waste/types'
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    #-------------------------------------------------------------------
    # Processing (new endpoint for NJ!)
    #-------------------------------------------------------------------

    # TODO: Implement:
    # - GET /processing/v1/{id}
    # - GET /processing/v1/active
    # - GET /processing/v1/inactive
    # - POST /processing/v1/createpackages
    # - POST /processing/v1/start
    # - POST /processing/v1/adjust
    # - PUT /processing/v1/finish
    # - PUT /processing/v1/unfinish
    # - DELETE /processing/v1/{id}
    # - GET /processing/v1/jobtypes/active
    # - GET /processing/v1/jobtypes/inactive
    # - POST /processing/v1/jobtypes
    # - PUT /processing/v1/jobtypes
    # - DELETE /processing/v1/jobtypes/{id}
    # - GET /processing/v1/jobtypes/attributes
    # - GET /processing/v1/jobtypes/categories


    #-------------------------------------------------------------------
    # Miscellaneous
    #-------------------------------------------------------------------


    # TODO: GET /caregivers/v1/status/{caregiverLicenseNumber}


    def get_units_of_measure(self, license_number=''):
        """Get all units of measurement.
        Args:
            license_number (str): A specific license number.
        Returns:
            (list): Returns a list of units of measure (dict).
        """
        url = METRC_UOM_URL
        params = self.format_params(license_number=license_number or self.primary_license)
        return self.request('get', url, params=params)


    def import_tags(self, file_path, row_start=0, row_end=None, number=10):
        """Import plant and package tags.
        Args:
            file_path (str): The file location of the tags.
            row_start (int): The row at which to begin importing tags.
            row_end (int): The row at which to end importing tags.
            number (int): The number of tags to import.
        Returns:
            (dict): Returns the tags as a dictionary.
        """
        if row_end:
            number = row_end - row_start
        df = read_excel(
            file_path,
            skiprows=1 + row_start,
            nrows=number,
            header=None,
            names=['tag', 'type', 'status', 'commissioned', 'used', 'detached']
        )
        return df.to_dict('records')



"""
Measurement Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Measurement data model.
"""
# Standard imports.
from dataclasses import dataclass, field
from typing import List

# Internal imports.
from .base import Model


@dataclass
class Measurement(Model):
    """A measurement is an amount measured by an analyst or scientific instrument.
    A calculation is applied to the measurement to get the final result."""
    _collection = 'organizations/%s/measurements'



"""
CoADoc | Parse CDX Analytics COAs
Copyright (c) 2022 Cannlytics

Authors:
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
    Keegan Skeate <https://github.com/keeganskeate>
Created: 11/25/2022
Updated: 11/25/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse CDX Analytics COA PDFs.

Data Points (✓):

    - analyses
    - methods
    - {analysis}_status
    - date_received
    - date_tested
    - lab_results_url
    - metrc_lab_id
    - metrc_source_id
    - producer
    - producer_street
    - producer_city
    - producer_state
    - producer_zipcode
    - producer_license_number
    - product_name
    - product_type
    - results
    - results_hash
    - sample_hash
    - total_cannabinoids
    - total_cbd (Calculated as total_cbd = cbd + 0.877 * cbda)
    - total_thc (Calculated as total_thc = delta_9_thc + 0.877 * thca)
    - total_terpenes

"""

# Standard imports.
from datetime import datetime
import json
import os
import re
from typing import Any, Optional

# External imports.
import pandas as pd
import pdfplumber
import requests

# Internal imports.
from cannlytics import __version__
from cannlytics.data.data import (
    create_hash,
    create_sample_id,
    find_first_value,
)
from cannlytics.data.web import download_google_drive_file
from cannlytics.utils import convert_to_numeric, snake_case
from cannlytics.utils.constants import DECARB


# === Tests ===
if __name__ == '__main__':

    from cannlytics.data.coas import CoADoc

    # Specify testing constants.
    doc = '../../../tests/assets/coas/cdx-analytics/Dosidos #22 - 120821DSDM22-5.pdf'
    parser = CoADoc()
    lims = parser.identify_lims(doc)

    report = pdfplumber.open(doc)
    for page in report.pages:
        text = page.extract_text()
        lines = text.split('\n')
        for line in lines:
            print(line)

            # Get sample details.

            # Get client details.

            # Get analyte results
            





"""
Statistics Module
Copyright (c) 2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 5/31/2022
Updated: 6/1/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>
"""
# Standard imports.
from datetime import datetime
import os
import shutil
from typing import Any, Optional

# External imports.
import pandas as pd
try:
    from sklearn.metrics import confusion_matrix
    import statsmodels.api as sm
except:
    pass

# Internal imports.
from cannlytics.firebase import (
    download_file,
    get_document,
    update_documents,
    upload_file,
)


def calculate_model_statistics(models, Y, X):
    """Determine prediction thresholds for a given model.
    Calculate a confusion matrix and returns prediction statistics.
    Args:
        models (list): A list of simultaneous prediction models.
        X (DataFrame): A DataFrame of explanatory variables.
        Y (DataFrame): A DataFrame of outcome variables.
    Returns:
        (dict): Returns a dictionary of statistics for each model.
    """
    X = X.loc[:, (X != 0).any(axis=0)]
    x = sm.add_constant(X)
    base, acc, fpr, fnr, tpr, tnr, info = {}, {}, {}, {}, {}, {}, {}
    for key in Y.columns:
        y = Y[key]
        y_bar = y.mean()
        model = models[key]
        if model:
            x_hat = x[list(model.params.keys())]
            y_hat = model.predict(x_hat)
            threshold = round(y_hat.quantile(1 - y_bar), 4)
            base[key] = threshold
            prediction = pd.Series(y_hat > threshold).astype(int)
            cm = confusion_matrix(y, prediction)
            tn, fp, fn, tp = cm.ravel()
            pos = sum(y)
            neg = len(y) - pos
            fpr[key] = round(fp / neg, 4)
            fnr[key] = round(fn / pos, 4)
            tpr[key] = round(tp / pos, 4)
            tnr[key] = round(tn / neg, 4)
            acc[key] = round((tp + tn) / (pos + neg), 4)
            info[key] = round((tp / pos) / (tn / neg), 4)
    stats = pd.DataFrame({
        'threshold': base,
        'false_positive_rate': fpr,
        'false_negative_rate': fnr,
        'true_positive_rate': tpr,
        'true_negative_rate': tnr,
        'accuracy': acc,
        'informedness': info,
    })
    stats = stats.fillna(0)
    return stats


def estimate_discrete_model(X, Y, method=None):
    """Estimate a prediction model(s) for discrete outcomes.
    The algorithm excludes all null columns, adds a constant,
    then fits probit model(s) for each effect variable.
    The user can specify their model, e.g. logit, probit, etc.
    Args:
        X (DataFrame): A DataFrame of explanatory variables.
        Y (DataFrame): A DataFrame of outcome variables.
        method (str, function): Specify 'probit', 'logit', or pass
            a statistical model of your choice with a `fit` method.
            A probit model is used by default (optional).
    Returns:
        (list): Returns a list of simultaneous prediction models.
    """
    X = X.loc[:, (X != 0).any(axis=0)]
    X = sm.add_constant(X)
    models = {}
    if method == 'logit':
        method = sm.Logit
    elif method is None or method == 'probit':
        method = sm.Probit
    for variable in Y.columns:
        try:
            y = Y[variable]
            model = method(y, X).fit(disp=0)
            models[variable] = model
        except:
            models[variable] = None # Error estimating!
    return models


def get_stats_model(
        ref: str,
        data_dir: Optional[str] = '/tmp',
        name: Optional[str] = None,
        bucket_name: Optional[str] = None,
    ):
    """Get a pre-built statistical model for use.
    First, gets the model data from Firebase Firestore.
    Second, downloads the pickle file and loads it into a model.
    Args:
        ref (str): The reference of the model data and file.
        data_dir (str): A folder to save the model files.
    Returns:
        (dict): Data about the model, including `model` and `model_stats`.
    """
    if name is None:
        name = ref.replace('/', '-')
    model_path = os.path.join(data_dir, name)
    if not os.path.exists(model_path):
        os.makedirs(model_path)
    data = get_document(ref)
    file_name = ref.split('/')[-1] + '.zip'
    zipped_file = os.path.join(data_dir, file_name)
    download_file(data['model_ref'], zipped_file, bucket_name)
    shutil.unpack_archive(zipped_file, model_path)
    models = {}
    for item in os.listdir(model_path):
        pickle_file = os.path.join(model_path, item)
        key = item.replace('model_', '').replace('.pickle', '')
        with open(pickle_file, 'rb') as f:
            models[key] = sm.load(f)
    data['model'] = models
    return data


def predict_stats_model(models, X, thresholds=None):
    """Predict outcomes for a given model and its thresholds.
    Add a constant column if necessary and only use model columns.
    Args:
        models (list): A list of simultaneous prediction models.
        X (DataFrame): A DataFrame of explanatory variables.
        thresholds (dict): A dictionary of thresholds to be used as
            decision rules for binary outcomes.
    Returns:
        (DataFrame): Returns predictions for each outcome variable.
    """
    x = X.assign(const=1)
    predictions = pd.DataFrame()
    for key, model in models.items():
        if not model:
            predictions[key] = 0
            continue
        x_hat = x[list(model.params.keys())]
        y_hat = model.predict(x_hat)
        if thresholds:
            threshold = thresholds[key]
            prediction = pd.Series(y_hat > threshold).astype(int)
            predictions[key] = prediction
        else:
            predictions[key] = y_hat
    return predictions


def upload_stats_model(
        models: Any,
        ref: str,
        name: Optional[str] = None,
        data_dir: Optional[str] = '/tmp',
        stats: Optional[Any] = None
    ):
    """Upload an statistical model for future use.
    Pickle each model, zip the model files, then upload the zipped file.
    Finally, record the file's data in Firebase Firestore.
    Args:
        models (dict): The list of effects models.
        ref (str): The reference for the model data and file.
        name (str): A name to save the model as (optional).
        data_dir (str): A directory to save the model files (optional).
        stats (DataFrame): Model summary statistics (optional).
    Returns:
        (dict): Returns the model data.
    """
    if name is None:
        name = ref.replace('/', '-')
    model_path = os.path.join(data_dir, name)
    if not os.path.exists(model_path):
        os.makedirs(model_path)
    if not isinstance(models, dict):
        models = {'model': models}
    for key, model in models.items():
        model_file = os.path.join(model_path, f'model_{key}.pickle')
        model.save(model_file)
    zipped_file = os.path.join(data_dir, name)
    shutil.make_archive(zipped_file, 'zip', model_path)
    file_ref = ref + '.zip'
    data = {
        'model_ref': file_ref,
        'model_stats': stats.to_dict(),
        'updated_at': datetime.now().isoformat(),
    }
    upload_file(file_ref, zipped_file + '.zip')
    update_documents([ref], [data])
    return data



"""
Data Market | Cannlytics
Copyright (c) 2021-2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 10/11/2021
Updated: 6/26/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

TODO:

    1. Refactor
        - Write docstrings.
        - Allow selling data as a pool.
        - Allow user to pass a Wallet instead of private_key.

    2. Publish assets metadata and associated services
        - Each asset is assigned a unique DID and a DID Document (DDO)
        - The DDO contains the asset's services including the metadata
        - The DID is registered on-chain with a URL of the metadata store
            to retrieve the DDO from
        `asset = ocean.assets.create(metadata, publisher_wallet)`

Read up:

    - https://www.quicknode.com/guides/web3-sdks/estimating-gas-price-using-pending-transactions-in-python

"""
# Internal packages.
from datetime import datetime
import os
import yaml

# External packages.
from dotenv import dotenv_values
try:
    from ocean_lib.ocean.ocean import Ocean
    from ocean_lib.web3_internal.wallet import Wallet
    from ocean_lib.web3_internal.currency import from_wei
except:
    pass

# External packages.
# try:
#     from ocean_lib.common.agreements.service_types import ServiceTypes
#     from ocean_lib.services.service import Service
#     from ocean_lib.data_provider.data_service_provider import DataServiceProvider
#     from ocean_lib.models.btoken import BToken # BToken is ERC20
#     from ocean_lib.ocean.ocean import Ocean
#     from ocean_lib.web3_internal.constants import ZERO_ADDRESS
#     from ocean_lib.web3_internal.currency import pretty_ether_and_wei, to_wei, from_wei
#     from ocean_lib.web3_internal.wallet import Wallet
# except:
#     # FIXME: ocean_lib doesn't like the App Engine environment.
#     pass


# def create_datatoken(wallet, nft, symbol, value):
#     """Create a datatoken related to a given NFT.
#     """
#     datatoken = nft.create_datatoken(value, symbol, from_wallet=wallet)
#     print(f"Created datatoken. Its address is {datatoken.address}")
#     return datatoken


# def publish_data_nft(ocean, wallet, symbol, value):
#     """Create Ocean instance, wallet, and publish an NFT token.
#     """
#     data_nft = ocean.create_data_nft(
#         value,
#         symbol,
#         wallet,
#         # token_uri='https://cannlytics.com/nft/',
#     )
#     print(f"Created data NFT. Its address is {data_nft.address}")
#     return data_nft



#-----------------------------------------------------------------------
# Test
#-----------------------------------------------------------------------

# Get the user's private key.
env_vars = dotenv_values('../../.env')
user_private_key = env_vars['TEST_PRIVATE_KEY1']
os.environ['ocean_network_url'] = env_vars['OCEAN_NETWORK_URL']
os.environ['ocean_network_url'] = env_vars['ADDRESS_FILE']
# =~/.ocean/ocean-contracts/artifacts/address.json

# Define data NFT value and symbol.


# Initialize the data ocean.
from ocean_lib.example_config import ExampleConfig
config = ExampleConfig.get_config()
ocean = Ocean(config)
# ocean = Ocean({
#     'address_file': '~/.ocean/ocean-contracts/artifacts/address.json',
#     'network_url': 'http://127.0.0.1:8545',
#     # 'network': 'https://rinkeby.infura.io/v3/1f0c7b5c219448f5a767a6ff0820d3e8',
#     # 'metadataCacheUri': 'https://aquarius.rinkeby.oceanprotocol.com',
#     # 'providerUri': 'https://provider.rinkeby.oceanprotocol.com',
# })

# Examine config.
print(f"config.network_url = '{ocean.config.network_url}'")
print(f"config.block_confirmations = {ocean.config.block_confirmations.value}")
print(f"config.metadata_cache_uri = '{ocean.config.metadata_cache_uri}'")
print(f"config.provider_url = '{ocean.config.provider_url}'")

# Get the user's wallet.
user_wallet = Wallet(
    ocean.web3,
    user_private_key,
    ocean.config.block_confirmations,
    ocean.config.transaction_timeout,
)
print(f"user_wallet.address = '{user_wallet.address}'")

# Check that the user has ETH.
assert user_wallet.web3.eth.get_balance(user_wallet.address) > 0, 'need ETH'

# Get OCEAN token balance.
print(f"Address of OCEAN token: {ocean.OCEAN_address}")
OCEAN_token = ocean.OCEAN_token
OCEAN_balance_in_wei = OCEAN_token.balanceOf(user_wallet.address)
OCEAN_balance_in_ether = from_wei(OCEAN_balance_in_wei)
print(f"Balance: {OCEAN_balance_in_ether} OCEAN")
if OCEAN_balance_in_wei == 0:
  print("WARNING: you don't have any OCEAN yet")

# TODO: Estimate gas price to display to user.

# Create a data NFT.
symbol = 'CAN-NFT-1'
value = '[{"test": "yes"}]'
data_nft = ocean.create_data_nft(
    value,
    symbol,
    user_wallet,
    owner=user_wallet.address,
    # token_uri='https://cannlytics.com/nft/',
)
print(f"Data NFT token name: {data_nft.token_name()}")
print(f"Data NFT token symbol: {data_nft.symbol()}")

# Create a datatoken for the data NFT.
symbol = 'CAN-DT-1'
value = 'Datatoken 1'
datatoken = data_nft.create_datatoken(value, symbol, user_wallet)
print(f"Datatoken name: {datatoken.token_name()}")
print(f"Datatoken symbol: {datatoken.symbol()}")

# Optional: Do both at the same time?
# ocean.create_nft_with_erc20()

# TODO: Save IDs / data to Firestore!!!
gas_price = from_wei(data_nft.get_gas_price(data_nft.web3))


# TODO: Get IDs from Firestore.
# - Initiate token?


#-----------------------------------------------------------------------
# Algorithm publishing and consumption.
#-----------------------------------------------------------------------

# Get a preview of what you will pay for an algorithm compute job.
# ocean.ocean_compute.get_c2d_environments(service.service_endpoint)
# ocean.retrieve_provider_fees_for_compute(datasets, algorithm_data, consumer_address, compute_environment, duration)


#-----------------------------------------------------------------------
# OLD v3
#-----------------------------------------------------------------------

# def get_wallet(ocean, private_key):
#     """Get a user's wallet given their private key.
#     Args:
#         ocean ():
#         private_key (str):
#     Returns:

#     """
#     return Wallet(ocean.web3, private_key, ocean.config.block_confirmations)


# def initialize_market(env_file='env.yaml'):
#     """Initialize an Ocean data marketplace.
#     Args:
#         env_file (str): An environment variable file,
#             `env.yaml` by default.
#     Returns:
    
#     """
#     config = None
#     with open(env_file, 'r') as f:
#         config = yaml.load(f, Loader=yaml.FullLoader)
#     os.environ['ocean_network_url'] = config['OCEAN_NETWORK_URL']
#     return Ocean(config)


# def publish_data(
#         ocean,
#         private_key,
#         files,
#         name,
#         symbol,
#         author,
#         data_license='CC0: Public Domain',
# ):
#     """Publish a dataset on the Ocean marketplace.
#     Publish metadata and service attributes on-chain.
#     The service urls will be encrypted before going on-chain.
#     They're only decrypted for datatoken owners upon consume.
#     Args:
#         ocean ():
#         private_key (str):
#         files (list):
#         name (str):
#         symbol (str):
#         author (str):
#         data_license (str): The license for the data,
#             `CC0: Public Domain` by default.
#     Returns:
#         ()
#     """
#     wallet = Wallet(ocean.web3, private_key, ocean.config.block_confirmations)
#     assert wallet.web3.eth.get_balance(wallet.address) > 0, 'need ETH'
#     print('Proceeding with wallet:', wallet.address)
#     data_token = ocean.create_data_token(name, symbol, wallet, blob=ocean.config.metadata_cache_uri)
#     # return data_token
#     token_address = data_token.address
#     print('Created token:', token_address)
#     date_created = datetime.now().isoformat()
#     metadata =  {
#         'main': {
#             'type': 'dataset',
#             'name': name,
#             'author': author,
#             'license': data_license,
#             'dateCreated': date_created,
#             'files': files,
#         }
#     }
#     service_attributes = {
#         'main': {
#             'name': 'dataAssetAccessServiceAgreement',
#             'creator': wallet.address,
#             'timeout': 3600 * 24,
#             'datePublished': date_created,
#             'cost': 1.0, # <don't change, this is obsolete>
#         }
#     }
#     service_endpoint = DataServiceProvider.get_url(ocean.config)
#     # FIXME:
#     download_service = Service(
#         service_endpoint=service_endpoint,
#         service_type=ServiceTypes.ASSET_ACCESS,
#         attributes=service_attributes,
#     )
#     assert wallet.web3.eth.get_balance(wallet.address) > 0, 'need ETH'
#     asset = ocean.assets.create(
#         metadata,
#         wallet,
#         # services=[download_service],
#         # service_descriptors=[],
#         data_token_address=token_address
#     )
#     print('Created asset:', asset.data_token_address)
#     assert token_address == asset.data_token_address
#     return data_token, asset


# def sell_data(
#         ocean,
#         private_key,
#         data_token,
#         amount,
#         fixed_price=True,
# ):
#     """Sell a dataset on the Ocean market.
#     Mint the datatokens.
#     In the create() step below, ganache OCEAN is needed.
#     Finally, Approve the datatoken for sale.
#     Args:
#         ocean ():
#         wallet ():
#         data_token ():
#         amount ():
#         fixed_price (bool): Whether or not to sell the data at a fixed price.
#     Returns:
#         (bool): Returns True if successful.
#     """
#     wallet = Wallet(ocean.web3, private_key, ocean.config.block_confirmations)
#     data_token.mint(wallet.address, to_wei(amount), wallet)
#     OCEAN_token = BToken(ocean.web3, ocean.OCEAN_address)
#     assert OCEAN_token.balanceOf(wallet.address) > 0, 'need OCEAN'
#     data_token.approve(
#         ocean.exchange._exchange_address,
#         to_wei(amount),
#         wallet
#     )
#     return True


# def buy_data(
#         ocean,
#         private_key,
#         token_address,
#         seller_wallet,
#         min_amount,
#         max_amount,
# ):
#     """Buy a dataset on the market.
#     Define wallet, verify that there is enough ganache ETH and OCEAN.
#     Create an exchange_id for a new exchange.
#     Args:
#         ocean ():
#         private_key (str):
#         token_address (str):
#         seller_wallet (Wallet):
#         min_amount (float):
#         max_amount (float):
#     Returns:

#     """
#     wallet = Wallet(ocean.web3, private_key, ocean.config.block_confirmations)
#     assert ocean.web3.eth.get_balance(wallet.address) > 0, 'need ganache ETH'
#     OCEAN_token = BToken(ocean.web3, ocean.OCEAN_address)
#     assert OCEAN_token.balanceOf(wallet.address) > 0, 'need ganache OCEAN'
#     exchange_id = ocean.exchange.create(token_address, to_wei(min_amount), seller_wallet)
#     tx_result = ocean.exchange.buy_at_fixed_rate(
#         to_wei(min_amount),
#         wallet,
#         to_wei(max_amount),
#         exchange_id,
#         token_address,
#         seller_wallet.address
#     )
#     assert tx_result, 'failed buying data tokens at fixed rate.'
#     # FIXME:
#     # print(f"Bob has {pretty_ether_and_wei(data_token.balanceOf(bob_wallet.address), data_token.symbol())}.")
#     # assert data_token.balanceOf(wallet.address) >= to_wei(1), "Bob didn't get 1.0 datatokens"



# def download_data(ocean, private_key, did):
#     """Download a dataset that is in a user's possession.
#     Points to the service object, send datatoken to the service,
#     and then downloads the dataset files. If the connection breaks,
#     then the request can be made again with the order_tx_id.
#     Args:
#         ocean ():
#         private_key ():
#         did (str): Dataset ID.
#     Returns:

#     """
#     wallet = Wallet(ocean.web3, private_key, ocean.config.block_confirmations)
#     fee_receiver = ZERO_ADDRESS # could also be market address
#     asset = ocean.assets.resolve(did)
#     service = asset.get_service(ServiceTypes.ASSET_ACCESS)
#     quote = ocean.assets.order(asset.did, wallet.address, service_index=service.index)
#     order_tx_id = ocean.assets.pay_for_service(
#         ocean.web3,
#         quote.amount,
#         quote.data_token_address,
#         asset.did,
#         service.index,
#         fee_receiver,
#         wallet,
#         service.get_c2d_address()
#     )
#     try:
#         file_path = ocean.assets.download(
#             asset.did,
#             service.index,
#             wallet,
#             order_tx_id,
#             destination='./'
#         )
#         return file_path
#     except:
#         return order_tx_id



"""
Metrc API URLs | Cannlytics Module
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>
"""

METRC_API_BASE_URL = 'https://api-%s.metrc.com'
METRC_API_BASE_URL_TEST = 'https://sandbox-api-%s.metrc.com'
METRC_BATCHES_URL = '/plantbatches/v1/%s'
METRC_EMPLOYEES_URL = '/employees/v1/'
METRC_FACILITIES_URL = '/facilities/v1/'
METRC_LOCATIONS_URL = '/locations/v1/%s'
METRC_HARVESTS_URL = '/harvests/v1/%s'
METRC_ITEMS_URL = '/items/v1/%s'
METRC_LAB_RESULTS_URL = '/labtests/v1/%s'
METRC_PACKAGES_URL = '/packages/v1/%s'
METRC_PATIENTS_URL = '/patients/v1/%s'
METRC_PLANTS_URL = '/plants/v1/%s'
METRC_RECEIPTS_URL = '/sales/v1/receipts/%s'
METRC_SALES_URL = '/sales/v1/%s'
METRC_STRAINS_URL = '/strains/v1/%s'
METRC_TRANSACTIONS_URL = '/sales/v1/transactions/%s'
METRC_TRANSFERS_URL = '/transfers/v1/%s'
METRC_TRANSFER_PACKAGES_URL = '/transfers/v1/delivery/%s/%s'
METRC_TRANSFER_TEMPLATE_URL = '/transfers/v1/templates/%s'
METRC_UOM_URL = '/unitsofmeasure/v1/active'
